"""
This module contains the connection handling.
"""
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine


def enable_foreign_keys(dbapi_conn):
    """
    Enable foreign key constraints for this connection.

    @param dbapi_conn: database connection
    @type dbapi_conn: L{sqlalchemy.engine.interfaces.DBAPIConnection}
    """
    cursor = dbapi_conn.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


@event.listens_for(Engine, "connect")
def enable_foreign_keys_on_connect(dbapi_connection, connection_record):
    """
    Enable foreign keys if a sqlite connection has been made.

    @param dbapi_connection: database connection
    @type dbapi_connection: L{sqlalchemy.engine.interfaces.DBAPIConnection}
    @param connection_record: database connection record info
    @type connection_record: L{sqlalchemy.pool.ConnectionPoolEntry}
    """
    # Check if the connection URL is SQLite
    if "sqlite" in str(connection_record.driver_connection):
        enable_foreign_keys(dbapi_connection)


class ConnectionConfig(object):
    """
    This class holds all information for establishing a database connection.

    @ivar url: sqlalchemy database url to connect to
    @type url: L{str}
    @ivar verbose: if nonzero, be verbose
    @type verbose: L{bool}
    """
    def __init__(self, url, verbose=False):
        """
        The default constructor.

        @param url: sqlalchemy database url to connect to
        @type url: L{str}
        @param verbose: if nonzero, be verbose
        @type verbose: L{bool}
        """
        assert isinstance(url, str)

        self.url = url
        self.verbose = verbose

    def connect(self):
        """
        Establish a database connection.

        @return: the connected sqlalchemy engine
        @rtype: L{sqlalchemy.engine.Engine}
        """
        if self.verbose:
            print("Connecting to database...")
        engine = create_engine(
            self.url,
            echo=self.verbose,
        )
        if self.verbose:
            print("Connected.")
        return engine



"""
Import logic.

@var POST_COLUMNS: keys from dataset to add to post model
@type POST_COLUMNS: L{list} of L{str}
@var POST_FILTERS: a dictionary mapping post columns to transformer to apply
@type POST_FILTERS: L{dict} of L{str} -> callable
@var COMMENT_COLUMNS: keys from dataset to add to comment model
@type COMMENT_COLUMNS: L{list} of L{str}
@var COMMENT_FILTERS: a dictionary mapping comment columns to transformer to apply
@type COMMENT_FILTERS: L{dict} of L{str} -> callable
"""
import json
import datetime

from sqlalchemy import select

from .jsonl import process_jsonl
from .db.models import Post, User, Comment, Subreddit, ARCTICZIM_USERNAME
from .util import chunked


POST_COLUMNS = [c.key for c in Post.__table__.columns]
POST_FILTERS = {
    "poll_data": json.dumps,
    "media_metadata": json.dumps,
    "edited": lambda x: {False: 0, True: -1}.get(x, x),
}
COMMENT_COLUMNS = [c.key for c in Comment.__table__.columns]
COMMENT_FILTERS = {
    "edited": lambda x: {False: 0, True: -1}.get(x, x)
}


def prepare_db(session):
    """
    Prepare a database.

    This inserts some default objects into the database.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    """
    # first, check if the user already exists
    if session.execute(select(User).where(User.name == ARCTICZIM_USERNAME)).one_or_none() is None:
        user = User(
            name=ARCTICZIM_USERNAME,
            created=datetime.datetime.now(),
        )
        session.add(user)
        session.commit()


def import_posts(session, posts):
    """
    Create a database post from a dataset post and insert it into the db.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}

    @param posts: list of dictionaries from dataset containing post data
    @type posts: L{list} of L{dict}
    """
    # get or create authors
    authors = {}
    for d in posts:
        author_name = d["author"]
        if author_name not in authors:
            author = session.get(User, author_name)
            if author is None:
                author = User(
                    # id=d["author_fullname"],
                    name=author_name,
                    created=datetime.datetime.fromtimestamp(
                        d["author_created_utc"]
                        if d.get("author_created_utc", None) is not None
                        else 0
                    ),
                )
                session.add(author)
            authors[author_name] = author
    # get or create subreddit
    subreddits = {}
    for d in posts:
        subreddit_name = d["subreddit"]
        if subreddit_name not in subreddits:
            subreddit = session.get(Subreddit, subreddit_name)
            if subreddit is None:
                subreddit = Subreddit(
                    # id=d["author_fullname"],
                    name=subreddit_name,
                    subscribers=d.get("subreddit_subscribers", 0),
                )
                session.add(subreddit)
            if d.get("subreddit_subscribers", 0) > subreddit.subscribers:
                subreddit.subscribers = d.get("subreddit_subscribers", 0)
            subreddits[subreddit_name] = subreddit
    # create posts
    post_objs = []
    for d in posts:
        # create kwargs
        kwargs = {}
        author_name = d["author"]
        subreddit_name = d["subreddit"]
        kwargs["author_name"] = author_name
        kwargs["author"] = authors[author_name]
        kwargs["subreddit_name"] = subreddit_name
        kwargs["subreddit"] = subreddits[subreddit_name]
        # fill in remaining kwargs
        for key in d.keys():
            if key in POST_COLUMNS:
                value = d[key]
                if key == "edited" and isinstance(value, bool):
                    if value:
                        value = -1
                    else:
                        value = 0
                kwargs[key] = POST_FILTERS.get(key, lambda x: x)(value)
        post = Post(**kwargs)
        rc = post.create_root_comment()
        rc.post = post
        post_objs.append(post)
        # generate a root comment
        session.add(post)
    session.commit()


def import_posts_from_file(session, path, batch_size=1000):
    """
    Import posts from a arcticshift dataset, adding them to the session.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param path: path to file to read
    @type path: L{str}
    @param batch_size: how many posts to import at once
    @type batch_size: L{int}
    """
    n = 0
    for post_batch in chunked(process_jsonl(path, desc="Importing posts"), batch_size):
        import_posts(session, post_batch)
        n += len(post_batch)
    print("Imported {} posts.".format(n))


def import_comments(session, comments):
    """
    Create database comments from dataset comments and insert them into the db.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param comments: list of dictionaries from dataset containing comment data
    @type comments: L{list} of L{dict}
    @return: the amount of failed imports
    @rtype: L{int}
    """
    n_fails = 0
    # get or create authors
    authors = {}
    for d in comments:
        author_name = d["author"]
        if author_name not in authors:
            author = session.get(User, author_name)
            if author is None:
                author = User(
                    # id=d["author_fullname"],
                    name=author_name,
                    created=datetime.datetime.fromtimestamp(
                        d["author_created_utc"]
                        if d.get("author_created_utc", None) is not None
                        else 0
                    ),
                )
                session.add(author)
            authors[author_name] = author
    # get or create subreddits
    subreddits = {}
    for d in comments:
        subreddit_name = d["subreddit"]
        if subreddit_name not in subreddits:
            subreddit = session.get(Subreddit, subreddit_name)
            if subreddit is None:
                subreddit = Subreddit(
                    # id=d["author_fullname"],
                    name=subreddit_name,
                    subscribers=d.get("subreddit_subscribers", 0),
                )
                session.add(subreddit)
            if d.get("subreddit_subscribers", 0) > subreddit.subscribers:
                subreddit.subscribers = d.get("subreddit_subscribers", 0)
            subreddits[subreddit_name] = subreddit
    # get posts
    posts = {}
    for d in comments:
        post_name = d["link_id"]
        if post_name not in posts:
            post = session.execute(select(Post).where(Post.name == post_name)).one_or_none()
            if post is not None:
                posts[post_name] = post[0]
    # create comments:
    parents = {}
    for d in comments:
        kwargs = {}
        author_name = d["author"]
        author = authors[author_name]
        kwargs["author"] = author
        kwargs["author_name"] = author_name
        subreddit_name = d["subreddit"]
        subreddit = subreddits[subreddit_name]
        kwargs["subreddit"] = subreddit
        kwargs["subreddit_name"] = subreddit_name
        post = posts.get(d["link_id"])
        if post is None:
            n_fails += 1
            continue
        kwargs["post"] = post
        parent_id = d["parent_id"]
        if parent_id not in parents:
            parent = session.execute(select(Comment).where(Comment.name == parent_id)).one_or_none()
            if parent is None:
                # can't insert comment before parent
                n_fails += 1
                continue
            parent = parent[0]
            parents[parent_id] = parent
        else:
            parent = parents[parent_id]
        kwargs["parent"] = parent
        # fill in remaining kwargs
        for key in d.keys():
            if key in COMMENT_COLUMNS:
                value = d[key]
                kwargs[key] = COMMENT_FILTERS.get(key, lambda x: x)(value)
        comment = Comment(**kwargs)
        session.add(comment)
        parents[comment.name] = comment
    session.commit()
    return n_fails


def import_comments_from_file(session, path, batch_size=1000):
    """
    Import comments from a arcticshift dataset, adding them to the session.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param path: path to file to read
    @type path: L{str}
    @param batch_size: how many comments to import at once
    @type batch_size: L{int}
    """
    n = 0
    n_fails = 0
    for comment_batch in chunked(process_jsonl(path, desc="Importing comments"), batch_size):
        cur_fails = import_comments(session, comment_batch)
        n += len(comment_batch) - cur_fails
        n_fails += cur_fails
    print("Imported {} comments, failed to import {} comments..".format(n, n_fails))



"""
The renderer generates HTML pages.

@var POSTS_PER_PAGE: how many posts are max allowed on a single page of a subreddit, ...
@type POSTS_PER_PAGE: L{int}
@var MAX_ITEMS_PER_RESULT: a subresult is submitted after this many items
@type MAX_ITEMS_PER_RESULT: L{int}
@var SUBREDDITS_ON_INDEX_PAGE: number of subreddits that should appear on the index page
@type SUBREDDITS_ON_INDEX_PAGE: L{int}
@var SUBREDDITS_PER_PAGE: how many subreddits should be listed on a page
@type SUBREDDITS_PER_PAGE: L{int}
"""
import json
import datetime
import math
import pprint

import htmlmin
import mistune
from jinja2 import Environment, PackageLoader, select_autoescape, Undefined

# optional optimization dependencies
try:
    import minify_html
except ImportError:
    minify_html = None

from ..util import format_size, format_number, get_resource_file_path, parse_reddit_url
from ..downloader import MediaFileManager
from .buckets import BucketMaker
from .custommistune import CustomMistuneBlockLevelParser, relative_url_plugin


POSTS_PER_PAGE = 20
MAX_ITEMS_PER_RESULT = 200
SUBREDDITS_ON_INDEX_PAGE = 20
SUBREDDITS_PER_PAGE = 40


class RenderedObject(object):
    """
    Base class for render results.
    """
    pass


class HtmlPage(RenderedObject):
    """
    This class holds the representation of a single rendered page.

    @ivar path: absolute path of the rendered page
    @type path: L{str}
    @ivar title: title of the page
    @type title: L{str}
    @ivar content: the HTML code of the page
    @type content: L{str}
    @ivar is_front: True if this is a front article
    @type is_front: L{bool}
    """
    def __init__(self, path, title, content, is_front=True):
        """
        The default constructor.

        @param path: absolute path of the rendered page
        @type path: L{str}
        @param title: title of the page
        @type title: L{str}
        @param content: the HTML code of the page
        @type content: L{str}
        @param is_front: True if this is a front article
        @type is_front: L{bool}
        """
        assert isinstance(path, str)
        assert isinstance(title, str)
        assert isinstance(content, str)
        assert isinstance(is_front, bool)
        self.path = path
        self.title = title
        self.content = content
        self.is_front = is_front

    def __str__(self):
        return "{}(path={}, title={}, content={} bytes)".format(self.__class__.__name__, self.path, self.title, len(self.content))


class JsonObject(RenderedObject):
    """
    This class holds a rendered json object.

    @ivar path: absolute path the object should be stored at
    @type path: L{str}
    @ivar title: title of the object
    @type title: L{str}
    @ivar content: the serialized json object to store
    @type content: any or L{str}
    """
    def __init__(self, path, title, content):
        """
        The default constructor.

        @ivar path: absolute path the object should be stored at
        @type path: L{str}
        @ivar title: title of the object
        @type title: L{str}
        @ivar content: the json object to store
        @type content: json-serializable
        """
        assert isinstance(path, str)
        assert isinstance(title, str)
        self.path = path
        self.title = title
        if isinstance(content, str):
            self.content = content
        else:
            self.content = json.dumps(content, separators=(",", ":"))

    def __str__(self):
        return "{}(path={}, title={}, content={} bytes)".format(self.__class__.__name__, self.path, self.title, len(self.content))


class Script(RenderedObject):
    """
    This class holds a rendered js script.

    @ivar path: absolute path the script should be stored at
    @type path: L{str}
    @ivar title: title of the object
    @type title: L{str}
    @ivar content: the script itself
    @type content: L{str}
    """
    def __init__(self, path, title, content):
        """
        The default constructor.

        @ivar path: absolute path the object should be stored at
        @type path: L{str}
        @ivar title: title of the script
        @type title: L{str}
        @ivar content: the script to store
        @type content: L{str}
        """
        assert isinstance(path, str)
        assert isinstance(title, str)
        self.path = path
        self.title = title
        self.content = content

    def __str__(self):
        return "{}(path={}, title={}, content={} bytes)".format(self.__class__.__name__, self.path, self.title, len(self.content))


class Redirect(RenderedObject):
    """
    This class holds redirect information.

    @ivar source: source path
    @type source: L{str}
    @ivar target: target path to redirect to
    @type target: L{str}
    @ivar title: title of the redirect
    @type title: L{str}
    @ivar is_front: True if this is a front article
    @type is_front: L{bool}
    """
    def __init__(self, source, target, title, is_front=False):
        """
        The default constructor.

        @param source: source path
        @type source: L{str}
        @param target: target path to redirect to
        @type target: L{str}
        @param title: title of the redirect
        @type title: L{str}
        @param is_front: True if this is a front article
        @type is_front: L{bool}
        """
        self.source = source
        self.target = target
        self.title = title
        self.is_front = is_front

    def __str__(self):
        return "{}(path={}, target={}, title={})".format(self.__class__.__name__, self.source, self.target, self.title)


class RenderResult(object):
    """
    This class encapsulates a list of  multiple L{RenderedObject},
    which together make up a rendered result.

    @ivar _objects: list of the rendered objects
    @type _objects: L{list} of L{RenderedObject}
    """
    def __init__(self, objects=None):
        """
        The default constructor.

        @param objects: list of the rendered objects or a rendered object, if any
        @type objects: L{None} or L{RenderedObject} or L{list} of L{RenderedObject}
        """
        if objects is None:
            self._objects = []
        elif isinstance(objects, RenderedObject):
            self._objects = [objects]
        elif isinstance(objects, (tuple, list)):
            self._objects = list(objects)
        else:
            raise TypeError("Expected None, RenderedObject or list/tuple of RenderedObject, got {} instead!".format(repr(objects)))

    def __str__(self):
        return "{}({})".format(self.__class__.__name__, pprint.pformat([str(o) for o in self.iter_objects()]))

    def add(self, obj):
        """
        Add an object to the result.

        @param obj: object to add
        @type obj: L{RenderedObject}
        """
        self._objects.append(obj)

    def iter_objects(self):
        """
        Iterate over the objects in this result.

        @yields: the objects in this result
        @ytype: L{RenderedObject}
        """
        for obj in self._objects:
            yield obj


class FileReferences(RenderedObject):
    """
    A list of MediaFile uids that were referenced in the post.

    @ivar uids: list of referenced uids
    @type uids: L{list} of L{int}
    """
    def __init__(self, uids):
        """
        The default constructor.

        @param uids: list of referenced uids
        @type uids: L{list} of L{int}
        """
        self.uids = uids


class SubredditInfo(object):
    """
    A helper class that holds details about a subreddit.

    @ivar name: name of this subreddit
    @type name: L{str}
    @ivar posts: number of posts in this subreddit
    @type posts: L{int}
    """
    def __init__(self, name, posts):
        """
        The default constructor.

        @param name: name of this subreddit
        @type name: L{str}
        @param posts: number of posts in this subreddit
        @type posts: L{int}
        """
        assert isinstance(name, str)
        assert isinstance(posts, int)
        self.name = name
        self.posts = posts


class RenderOptions(object):
    """
    Options for the renderer.

    @ivar with_stats: if nonzero, include statistics
    @type with_stats: L{bool}
    @ivar with_users: if nonzero, include user pages<
    @type with_users: L{bool}
    @ivar with_videos: if nonzero, include videos
    @type with_videos: L{bool}
    """
    def __init__(self, with_stats=True, with_users=True, with_videos=False):
        """
        The default constructor.

        @param with_stats: if nonzero, include statistics
        @type with_stats: L{bool}
        @param with_users: if nonzero, include user pages<
        @type with_users: L{bool}
        @param with_videos: if nonzero, include videos
        @type with_videos: L{bool}
        """
        self.with_stats = with_stats
        self.with_users = with_users
        self.with_videos = with_videos


class HtmlRenderer(object):
    """
    The HTML renderer renders HTML pages for various objects.

    @cvar MISTUNE_PLUGINS: list of plugins to use for mistune
    @type MISTUNE_PLUGINS: L{list} of L{str}

    @ivar worker: worker this renderer is for
    @type worker: L{arcticzim.zimbuild.worker.Worker}
    @ivar environment: the jinja2 environment used to render templates
    @type environment: L{jinja2.Environment}
    @ivar options: render options
    @type options: L{RenderOptions}
    @ivar filemanager: the media file manager
    @type filemanager: L{arcticzim.downloader.MediaFileManager}
    @ivar reference_rewrite: the URL rewriter for reddit references
    @type reference_rewriter: L{arcticzim.fetcher.ReferenceUrlRewriter}
    @ivar markdown: mistune markdown renderer
    @type markdown: L{mistune.markdown.Markdown}
    """

    MISTUNE_PLUGINS = [
        "strikethrough",
        "footnotes",
        "table",
        "url",
        "superscript",
        "subscript",
        "spoiler",
    ]

    def __init__(self, worker, options, filemanager, reference_rewriter):
        """
        The default constructor.

        @param worker: worker this renderer is for
        @type worker: L{arcticzim.zimbuild.worker.Worker}
        @param options: render options
        @type options: L{RenderOptions}
        @param filemanager: the media file manager
        @type filemanager: L{arcticzim.downloader.MediaFileManager}
        @param reference_rewriter: the URL rewriter for reddit references
        @type reference_rewriter: L{arcticzim.fetcher.ReferenceUrlRewriter}
        """
        assert isinstance(options, RenderOptions)
        assert isinstance(filemanager, MediaFileManager)
        self.worker = worker
        self.options = options
        self.filemanager = filemanager
        self.reference_rewriter = reference_rewriter

        # setup mistune (markdown renderer)
        self.markdown = mistune.create_markdown(plugins=self.MISTUNE_PLUGINS)
        self.markdown.block = CustomMistuneBlockLevelParser()
        relative_url_plugin(self.markdown)

        # setup jinja environment
        self.environment = Environment(
            loader=PackageLoader("arcticzim.zimbuild"),
            auto_reload=False,
            autoescape=select_autoescape(),
        )

        # configure environment globals
        self.environment.globals["options"] = self.options

        # configure filters
        self.environment.filters["render_comment_text"] = self._render_comment_text_filter
        self.environment.filters["render_post_text"] = self._render_comment_text_filter
        self.environment.filters["render_wiki_text"] = self._render_comment_text_filter
        self.environment.filters["render_license"] = self._render_license_text_filter
        self.environment.filters["format_number"] = format_number
        self.environment.filters["format_size"] = format_size
        self.environment.filters["format_date"] = self._format_date
        self.environment.filters["format_timestamp"] = self._format_timestamp
        self.environment.filters["first_elements"] = self._first_elements
        self.environment.filters["default_index"] = self._default_index
        self.environment.filters["rewrite_url"] = self._rewrite_url_filter
        self.environment.filters["load_json"] = json.loads
        self.environment.filters["first_nonzero"] = self._first_nonzero_filter
        self.environment.filters["stickysort"] = self._stickysort_filter
        self.environment.filters["render_postsummary_by_url"] = self._render_postsummary_by_url
        self.environment.filters["debug"] = print

        # configure tests
        self.environment.tests["date"] = self._is_date
        self.environment.tests["local_post_url"] = self._is_local_post_url

    @staticmethod
    def minify_html(s):
        """
        Minify html code.

        @param s: html code to minify
        @type s: L{str}
        @return: the minified html
        @rtype: L{str}
        """
        if minify_html is None:
            # fall back to htmlmin
            return htmlmin.minify(
                s,
                remove_comments=True,
                remove_empty_space=True,
                reduce_boolean_attributes=True,
                # remove_optional_attribute_quotes=True,
                remove_optional_attribute_quotes=False,  # firefox complains for some tags
            )
        else:
            return minify_html.minify(
                s,
            )

    def _add_file_references(self, result):
        """
        Add file references for the observed files into the render results.

        @param result: result to add file references to
        @type result: L{RenderResult}
        """
        fro = FileReferences(list(set(self.filemanager.referenced_files)))
        result.add(fro)

    def render_post(self, post):
        """
        Render a post.

        @param post: post to render
        @type post: L{arcticzim.db.models.Post}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        self.filemanager.reset()
        result = RenderResult()
        post_template = self.environment.get_template("postpage.html.jinja")
        post_page = post_template.render(
            post=post,
            to_root="../../..",
        )
        result.add(
            HtmlPage(
                path="r/{}/{}/".format(post.subreddit.name, post.id),
                title=post.title,
                content=self.minify_html(post_page),
                is_front=True,
            ),
        )
        # add redirect from id -> post
        result.add(
            Redirect(
                post.id,
                "r/{}/{}/".format(post.subreddit.name, post.id),
                title=post.title,
                is_front=False,
            ),
        )
        # add poll data if necessary
        if post.is_poll:
            raw_poll_data = json.loads(post.poll_data)
            if not raw_poll_data:
                # for some reason, post.poll_data can be false
                poll_data = False
            elif ("vote_count" not in raw_poll_data["options"][0]):
                # if the poll data is for some reason not available, this happens
                poll_data = False
            else:
                poll_data = {
                    "labels": [option["text"] for option in raw_poll_data["options"]],
                    "datasets": [
                        {
                            "data": [option["vote_count"] for option in raw_poll_data["options"]],
                        }
                    ],
                }
            result.add(
                JsonObject(
                    path="r/{}/{}/poll.json".format(post.subreddit.name, post.id),
                    title="Poll data for {}".format(post.id),
                    content=poll_data,
                )
            )
        # add referenced files
        self._add_file_references(result)
        return result

    def directly_render_post_summary(self, post, to_root):
        """
        Render a post summary, directly returning the HTML.

        @param post: post to render
        @type post: L{arcticzim.db.models.Post}
        @param to_root: rootification prefix (see templates)
        @type to_root: L{str}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        summary_template = self.environment.get_template("postsummary.html.jinja")
        post_page = summary_template.render(
            post=post,
            with_subreddit=True,
            with_util_links=True,
            to_root=to_root,
        )
        return post_page

    def render_subreddit(self, subreddit, sort="top", posts=None, num_posts=None):
        """
        Render a subreddit.

        If posts is specified, it should be an iterable yielding lists
        of posts sorted by the required order. If it is
        not specified, it will be generated from subreddit.posts

        @param subreddit: subreddit to render
        @type subreddit: L{arcticzim.db.models.Subreddit}
        @param sort: by which metric posts should be sorted
        @type sort: L{str}
        @param posts: an iterable yielding the posts in the subreddit in a sorted order as described above
        @type posts: iterable yielding L{arcticzim.db.models.Post} or L{None}
        @param num_posts: number of posts in this subreddit
        @type num_posts: L{int} or L{None}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        # general preparations
        if posts is None:
            num_posts = len(subreddit.posts)
            if sort == "top":
                posts = sorted(subreddit.posts, key=lambda x: x.score, reverse=True)
            elif sort == "new":
                posts = sorted(subreddit.posts, key=lambda x: x.created_utc, reverse=True)
            else:
                raise Exception("Invalid subreddit post sort order!")
        else:
            assert num_posts is not None
        has_wiki = (len(subreddit.wikipages) > 0)
        result = RenderResult()
        items_in_result = 0
        if num_posts == 0:
            # subreddit has no posts
            # we do not return any page in this case
            yield result
            return
        if sort == "top":
            # by default, redirect to top page 1
            result.add(
                Redirect(
                    "r/{}/".format(subreddit.name),
                    "r/{}/top_page_1/".format(subreddit.name),
                    title="r/{}".format(subreddit.name),
                    is_front=True,
                ),
            )
            items_in_result += 1
        # prepare rendering the subreddit pages
        list_page_template = self.environment.get_template("subredditpage.html.jinja")
        num_pages = math.ceil(num_posts / POSTS_PER_PAGE)
        bucketmaker = BucketMaker(POSTS_PER_PAGE)
        # render the subreddit post list pages
        page_index = 1
        for post in posts:
            bucket = bucketmaker.feed(post)
            if bucket is not None:
                items_in_result += self._render_subreddit_postlist_page(
                    subreddit=subreddit,
                    posts=bucket,
                    page_index=page_index,
                    num_pages=num_pages,
                    template=list_page_template,
                    sort=sort,
                    has_wiki=has_wiki,
                    result=result,
                )
                if items_in_result >= MAX_ITEMS_PER_RESULT:
                    yield result
                    result = RenderResult()
                    items_in_result = 0
                page_index += 1
        bucket = bucketmaker.finish()
        if bucket is not None:
            items_in_result += self._render_subreddit_postlist_page(
                subreddit=subreddit,
                posts=bucket,
                page_index=page_index,
                num_pages=num_pages,
                template=list_page_template,
                has_wiki=has_wiki,
                sort=sort,
                result=result,
            )
            if items_in_result >= MAX_ITEMS_PER_RESULT:
                yield result
                result = RenderResult()
                items_in_result = 0
        yield result

    def _render_subreddit_postlist_page(self, subreddit, posts, page_index, num_pages, template, sort, has_wiki, result):
        """
        Helper function for rendering a list page of posts in a subreddit.

        This function renders a page of posts in the subreddit and adds the
        rendered page to the result.

        @param subreddit: subreddit this page is for
        @type subreddit: L{arcticzim.db.models.Subreddit}
        @param posts: list of posts that should be listed on this page
        @type posts: L{list} of L{arcticzim.db.models.Post}
        @param page_index: index of current page (1-based)
        @type page_index: L{int}
        @param num_pages: total number of pages
        @type num_pages: L{int}
        @param template: template that should be rendered
        @type template: L{jinja2.Template}
        @param sort: sort order this page is for
        @type sort: L{str}
        @param has_wiki: whether the subreddit has a wiki or not
        @type has_wiki: L{bool}
        @param result: result the rendered page should be added to
        @type result: L{RenderResult}
        @return: the number of items added to the render result
        @rtype: L{int}
        """
        page = template.render(
            to_root="../../..",
            subreddit=subreddit,
            posts=posts,
            num_pages=num_pages,
            cur_page=page_index,
            has_wiki=has_wiki,
            sort=sort,
        )
        result.add(
            HtmlPage(
                path="r/{}/{}_page_{}/".format(subreddit.name, sort, page_index),
                content=self.minify_html(page),
                title="r/{} - {} - Page {}".format(subreddit.name, sort, page_index),
                is_front=False,
            ),
        )
        return 1  # 1 item added

    def render_subreddit_stats(self, subreddit, stats):
        """
        Render the statistics page of a subreddit.

        @param subreddit: subreddit to render
        @type subreddit: L{arcticzim.db.models.Subreddit}
        @param stats: stats to render
        @type stats: L{arcticzim.zimbuild.statistics.PostListStatistics}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        template = self.environment.get_template("subredditstatspage.html.jinja")
        page = template.render(
            to_root="../../..",
            subreddit=subreddit,
            stats=stats,
            has_wiki=(len(subreddit.wikipages) > 0),
        )
        result.add(
            HtmlPage(
                path="r/{}/statistics/".format(subreddit.name),
                content=self.minify_html(page),
                title="r/{} - Staticstics".format(subreddit.name),
                is_front=True,
            ),
        )
        return result

    def render_subreddit_wiki(self, subreddit):
        """
        Render the statistics page of a subreddit.

        @param subreddit: subreddit to render
        @type subreddit: L{arcticzim.db.models.Subreddit}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        template = self.environment.get_template("subredditwikipage.html.jinja")
        for wikipage in subreddit.wikipages:
            basepath = wikipage.basepath
            to_root = "../../.." + ("/.." * basepath.count("/"))
            page = template.render(
                to_root=to_root,
                subreddit=subreddit,
                wikipage=wikipage,
                has_wiki=True,
            )
            result.add(
                HtmlPage(
                    path="r/{}/wiki/{}".format(subreddit.name, wikipage.basepath),
                    content=self.minify_html(page),
                    title="r/{} - Wiki - {}".format(subreddit.name, wikipage.basepath),
                    is_front=True,
                ),
            )
        index_template = self.environment.get_template("subredditwikiindexpage.html.jinja")
        page = index_template.render(
            to_root="../../..",
            subreddit=subreddit,
            has_wiki=True,
        )
        result.add(
            HtmlPage(
                path="r/{}/wiki/_pageindex".format(subreddit.name),
                content=self.minify_html(page),
                title="r/{} - Wiki - Index".format(subreddit.name),
                is_front=True,
            ),
        )
        # default wiki redirect
        result.add(
            Redirect(
                source="r/{}/wiki/".format(subreddit.name),
                target="r/{}/wiki/_pageindex".format(subreddit.name),
                title="r/{} - Wiki".format(subreddit.name),
                is_front=True,
            )
        )
        return result

    def render_subreddit_rules(self, subreddit):
        """
        Render the rules page of a subreddit.

        @param subreddit: subreddit to render
        @type subreddit: L{arcticzim.db.models.Subreddit}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        template = self.environment.get_template("subredditrulespage.html.jinja")
        page = template.render(
            to_root="../../..",
            subreddit=subreddit,
            has_wiki=(len(subreddit.wikipages) > 0),
        )
        result.add(
            HtmlPage(
                path="r/{}/rules/".format(subreddit.name),
                content=self.minify_html(page),
                title="r/{} - Rules".format(subreddit.name),
                is_front=True,
            ),
        )
        return result

    def render_user_posts(self, user, sort="top", posts=None, num_posts=None):
        """
        Render the posts of a user.

        If posts is specified, it should be an iterable yielding lists
        of posts sorted by the required order. If it is
        not specified, it will be generated from user.posts

        @param user: user whose posts to render
        @type user: L{arcticzim.db.models.User}
        @param sort: by which metric posts should be sorted
        @type sort: L{str}
        @param posts: an iterable yielding the posts made by this user in a sorted order as described above
        @type posts: iterable yielding L{arcticzim.db.models.Post} or L{None}
        @param num_posts: number of posts made by this user
        @type num_posts: L{int} or L{None}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """

        # general preparations
        if posts is None:
            num_posts = len(user.posts)
            if sort == "top":
                posts = sorted(user.posts, key=lambda x: x.score, reverse=True)
            elif sort == "new":
                posts = sorted(user.posts, key=lambda x: x.created_utc, reverse=True)
            else:
                raise Exception("Invalid subreddit post sort order!")
        else:
            assert num_posts is not None
        result = RenderResult()
        items_in_result = 0
        if sort == "top":
            # by default, redirect to top posts page 1
            result.add(
                Redirect(
                    "u/{}/".format(user.name),
                    "u/{}/posts_top_page_1/".format(user.name),
                    title="u/{}".format(user.name),
                    is_front=True,
                ),
            )
            items_in_result += 1
        if num_posts == 0:
            # user has no posts
            empty_template = self.environment.get_template("useremptypage.html.jinja")
            page = empty_template.render(
                to_root="../../..",
                user=user,
                part="posts",
            )
            result.add(
                HtmlPage(
                    path="u/{}/posts_{}_page_1/".format(user.name, sort),
                    content=self.minify_html(page),
                    title="u/{} - Posts".format(user.name),
                    is_front=False,
                ),
            )
            yield result
            return
        # prepare rendering the user pages
        list_page_template = self.environment.get_template("userpostspage.html.jinja")
        num_pages = math.ceil(num_posts / POSTS_PER_PAGE)
        bucketmaker = BucketMaker(POSTS_PER_PAGE)
        # render the user post list pages
        page_index = 1
        for post in posts:
            bucket = bucketmaker.feed(post)
            if bucket is not None:
                items_in_result += self._render_user_postlist_page(
                    user=user,
                    posts=bucket,
                    page_index=page_index,
                    num_pages=num_pages,
                    template=list_page_template,
                    sort=sort,
                    result=result,
                )
                if items_in_result >= MAX_ITEMS_PER_RESULT:
                    yield result
                    result = RenderResult()
                    items_in_result = 0
                page_index += 1
        bucket = bucketmaker.finish()
        if bucket is not None:
            items_in_result += self._render_user_postlist_page(
                user=user,
                posts=bucket,
                page_index=page_index,
                num_pages=num_pages,
                template=list_page_template,
                sort=sort,
                result=result,
            )
            if items_in_result >= MAX_ITEMS_PER_RESULT:
                yield result
                result = RenderResult()
                items_in_result = 0
        yield result

    def _render_user_postlist_page(self, user, posts, page_index, num_pages, template, sort, result):
        """
        Helper function for rendering a list page of posts made by a user.

        This function renders a page of posts made by the user and adds the
        rendered page to the result.

        @param user: user this page is for
        @type user: L{arcticzim.db.models.User}
        @param posts: list of posts that should be listed on this page
        @type posts: L{list} of L{arcticzim.db.models.Post}
        @param page_index: index of current page (1-based)
        @type page_index: L{int}
        @param num_pages: total number of pages
        @type num_pages: L{int}
        @param template: template that should be rendered
        @type template: L{jinja2.Template}
        @param sort: sort order this page is for
        @type sort: L{str}
        @param result: result the rendered page should be added to
        @type result: L{RenderResult}
        @return: the number of items added to the render result
        @rtype: L{int}
        """
        page = template.render(
            to_root="../../..",
            user=user,
            posts=posts,
            num_pages=num_pages,
            cur_page=page_index,
            sort=sort,
        )
        result.add(
            HtmlPage(
                path="u/{}/posts_{}_page_{}/".format(user.name, sort, page_index),
                content=self.minify_html(page),
                title="u/{} - {} Posts - Page {}".format(user.name, sort, page_index),
                is_front=False,
            ),
        )
        return 1  # 1 item added

    def render_user_comments(self, user, sort="top", comments=None, num_comments=None):
        """
        Render the comments of a user.

        If comments is specified, it should be an iterable yielding lists
        of comments sorted by the required order. If it is
        not specified, it will be generated from user.comments

        @param user: user whose comments to render
        @type user: L{arcticzim.db.models.User}
        @param sort: by which metric posts should be sorted
        @type sort: L{str}
        @param comments: an iterable yielding the comments made by the user in a sorted order as described above
        @type comments: iterable yielding L{arcticzim.db.models.Comment} or L{None}
        @param num_comments: number of comments made by this user
        @type num_comments: L{int} or L{None}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """

        # general preparations
        if comments is None:
            num_comments = len(user.comments)
            if sort == "top":
                comments = sorted(user.comments, key=lambda x: x.score, reverse=True)
            elif sort == "new":
                comments = sorted(user.comments, key=lambda x: x.created_utc, reverse=True)
            else:
                raise Exception("Invalid subreddit comment sort order!")
        else:
            assert num_comments is not None
        result = RenderResult()
        items_in_result = 0
        if num_comments == 0:
            # user has no comments
            empty_template = self.environment.get_template("useremptypage.html.jinja")
            page = empty_template.render(
                to_root="../../..",
                user=user,
                part="comments",
            )
            result.add(
                HtmlPage(
                    path="u/{}/comments_{}_page_1/".format(user.name, sort),
                    content=self.minify_html(page),
                    title="u/{} - Comments".format(user.name),
                    is_front=False,
                ),
            )
            yield result
            return
        # prepare rendering the user pages
        list_page_template = self.environment.get_template("usercommentspage.html.jinja")
        num_pages = math.ceil(num_comments / POSTS_PER_PAGE)
        bucketmaker = BucketMaker(POSTS_PER_PAGE)
        # render the user post list pages
        page_index = 1
        for comment in comments:
            bucket = bucketmaker.feed(comment)
            if bucket is not None:
                items_in_result += self._render_user_commentslist_page(
                    user=user,
                    comments=bucket,
                    page_index=page_index,
                    num_pages=num_pages,
                    template=list_page_template,
                    sort=sort,
                    result=result,
                )
                if items_in_result >= MAX_ITEMS_PER_RESULT:
                    yield result
                    result = RenderResult()
                    items_in_result = 0
                page_index += 1
        bucket = bucketmaker.finish()
        if bucket is not None:
            items_in_result += self._render_user_commentslist_page(
                user=user,
                comments=bucket,
                page_index=page_index,
                num_pages=num_pages,
                template=list_page_template,
                sort=sort,
                result=result,
            )
            if items_in_result >= MAX_ITEMS_PER_RESULT:
                yield result
                result = RenderResult()
                items_in_result = 0
        yield result

    def _render_user_commentslist_page(self, user, comments, page_index, num_pages, template, sort, result):
        """
        Helper function for rendering a list page of comments made by a user.

        This function renders a page of comments made by the user and adds the
        rendered page to the result.

        @param user: user this page is for
        @type user: L{arcticzim.db.models.User}
        @param comments: list of comments that should be listed on this page
        @type comments: L{list} of L{arcticzim.db.models.Comment}
        @param page_index: index of current page (1-based)
        @type page_index: L{int}
        @param num_pages: total number of pages
        @type num_pages: L{int}
        @param template: template that should be rendered
        @type template: L{jinja2.Template}
        @param sort: sort order this page is for
        @type sort: L{str}
        @param result: result the rendered page should be added to
        @type result: L{RenderResult}
        @return: the number of items added to the render result
        @rtype: L{int}
        """
        page = template.render(
            to_root="../../..",
            user=user,
            comments=comments,
            num_pages=num_pages,
            cur_page=page_index,
            sort=sort,
        )
        result.add(
            HtmlPage(
                path="u/{}/comments_{}_page_{}/".format(user.name, sort, page_index),
                content=self.minify_html(page),
                title="u/{} - {} Comments - Page {}".format(user.name, sort, page_index),
                is_front=False,
            ),
        )
        return 1  # 1 item added

    def render_user_stats(self, user, stats):
        """
        Render the statistics page of a user.

        @param user: user to render
        @type user: L{arcticzim.db.models.User}
        @param stats: stats to render
        @type stats: L{arcticzim.zimbuild.statistics.PostListStatistics}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        template = self.environment.get_template("userstatspage.html.jinja")
        page = template.render(
            to_root="../../..",
            user=user,
            stats=stats,
        )
        result.add(
            HtmlPage(
                path="u/{}/statistics/".format(user.name),
                content=self.minify_html(page),
                title="u/{} - Staticstics".format(user.name),
                is_front=True,
            ),
        )
        return result

    def render_index(self, subreddit_infos):
        """
        Render the index page.

        @param subreddit_infos: a list of subreddits that should be listed on the index page
        @type subreddit_infos: L{list} of L{SubredditInfo}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        template = self.environment.get_template("indexpage.html.jinja")
        page = template.render(
            to_root=".",
            subreddit_infos=subreddit_infos,
        )
        result.add(
            HtmlPage(
                path="index.html",
                title="Welcome to ArcticZim!",
                content=self.minify_html(page),
                is_front=True,
            ),
        )
        return result

    def render_subreddit_list(self, subreddit_infos):
        """
        Render the list of subreddits.

        @param subreddit_infos: a list of all subreddits, sorted.
        @type subreddit_infos: L{list} of L{SubredditInfo}
        @yields: the generated pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        result.add(
            Redirect(
                "subreddits/",
                "subreddits/subreddits_page_1/",
                title="All Subreddits",
                is_front=True,
            ),
        )
        list_page_template = self.environment.get_template("subredditlistpage.html.jinja")
        num_subreddits = len(subreddit_infos)
        num_pages = math.ceil(num_subreddits / SUBREDDITS_PER_PAGE)
        bucketmaker = BucketMaker(SUBREDDITS_PER_PAGE)
        # render the user post list pages
        page_index = 1
        items_in_result = 0
        for subreddit_info in subreddit_infos:
            bucket = bucketmaker.feed(subreddit_info)
            if bucket is not None:
                items_in_result += self._render_subreddit_list_page(
                    subreddit_infos=bucket,
                    page_index=page_index,
                    num_pages=num_pages,
                    template=list_page_template,
                    result=result,
                )
                if items_in_result >= MAX_ITEMS_PER_RESULT:
                    yield result
                    result = RenderResult()
                    items_in_result = 0
                page_index += 1
        bucket = bucketmaker.finish()
        if bucket is not None:
            items_in_result += self._render_subreddit_list_page(
                subreddit_infos=bucket,
                page_index=page_index,
                num_pages=num_pages,
                template=list_page_template,
                result=result,
            )
            if items_in_result >= MAX_ITEMS_PER_RESULT:
                yield result
                result = RenderResult()
                items_in_result = 0
        yield result

    def _render_subreddit_list_page(self, subreddit_infos, page_index, num_pages, template, result):
        """
        Helper function for rendering a list page of the subreddit list.

        This function renders a page of subreddit infos and adds the
        rendered page to the result.

        @param subreddit_infos: list of subreddit_infos that should be listed on this page
        @type subreddit_infos: L{list} of L{SubredditInfo}
        @param page_index: index of current page (1-based)
        @type page_index: L{int}
        @param num_pages: total number of pages
        @type num_pages: L{int}
        @param template: template that should be rendered
        @type template: L{jinja2.Template}
        @param result: result the rendered page should be added to
        @type result: L{RenderResult}
        @return: the number of items added to the render result
        @rtype: L{int}
        """
        page = template.render(
            to_root="../..",
            subreddit_infos=subreddit_infos,
            num_pages=num_pages,
            cur_page=page_index,
        )
        result.add(
            HtmlPage(
                path="subreddits/subreddits_page_{}/".format(page_index),
                content=self.minify_html(page),
                title="All Subreddits - Page {}".format(page_index),
                is_front=False,
            ),
        )
        return 1  # 1 item added

    def render_scripts(self):
        """
        Generate the js scripts.

        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        # jQuery
        path = get_resource_file_path("jquery", "jquery.js")
        with open(path, "r", encoding="utf-8") as fin:
            script = fin.read()
        result.add(
            Script(
                path="scripts/jquery.js",
                content=script,
                title="jquery.js",
            ),
        )
        # chartjs
        path = get_resource_file_path("chartjs", "chart.js")
        with open(path, "r", encoding="utf-8") as fin:
            script = fin.read()
        result.add(
            Script(
                path="scripts/chart.js",
                content=script,
                title="Chart.js",
            ),
        )
        # polls
        path = get_resource_file_path("poll.js")
        with open(path, "r", encoding="utf-8") as fin:
            script = fin.read()
        result.add(
            Script(
                path="scripts/poll.js",
                content=script,
                title="Poll.js",
            ),
        )
        # collapser
        path = get_resource_file_path("collapser.js")
        with open(path, "r", encoding="utf-8") as fin:
            script = fin.read()
        result.add(
            Script(
                path="scripts/collapser.js",
                content=script,
                title="Collapser.js",
            ),
        )
        # preview
        path = get_resource_file_path("preview.js")
        with open(path, "r", encoding="utf-8") as fin:
            script = fin.read()
        result.add(
            Script(
                path="scripts/preview.js",
                content=script,
                title="preview.js",
            ),
        )
        return result

    def render_info_pages(self):
        """
        Render the various information pages.

        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        template = self.environment.get_template("infopage.html.jinja")
        info_page = template.render(
            to_root="..",
        )
        result.add(
            HtmlPage(
                path="info/",
                title="ArcticZim - Informations",
                content=self.minify_html(info_page),
                is_front=True,
            ),
        )
        license_template = self.environment.get_template("licensepage.html.jinja")
        license_paths = {
            "jQuery": ("jquery", "LICENSE.txt"),
            "Chart.js": ("chartjs", "LICENSE.md"),
        }
        licenses = {
            k: open(get_resource_file_path(*v)).read()
            for k, v in license_paths.items()
        }
        licenses["Icons"] = 'Several icons have been taken (or use assets from) [iconsDB.com](https://iconsdb.com).'
        license_page = license_template.render(
            to_root="../..",
            licenses=licenses,
        )
        result.add(
            HtmlPage(
                path="info/licenses/",
                title="ArcticZim - Licenses",
                content=self.minify_html(license_page),
                is_front=True,
            ),
        )
        return result

    def render_global_stats(self, stats):
        """
        Render the global statistics.

        @param stats: stats to render
        @type stats: L{arcticzim.zimbuild.statistics.PostListStatistics}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        result = RenderResult()
        template = self.environment.get_template("globalstatspage.html.jinja")
        page = template.render(
            to_root="..",
            stats=stats,
        )
        result.add(
            HtmlPage(
                path="statistics/",
                content=self.minify_html(page),
                title="ArcticZim - Staticstics",
                is_front=True,
            ),
        )
        return result

    # =========== filters ===============

    def _rewrite_url_filter(self, value, to_root):
        """
        Rewrite a URL.

        @param value: URL to rewrite
        @type value: L{str}
        @param to_root: same as in the templates
        @type to_root: L{str}
        """
        value = self.filemanager.rewrite_url(value, to_root=to_root)
        value = self.reference_rewriter.rewrite_url(value, to_root=to_root)
        return value

    def _render_comment_text_filter(self, value, to_root):
        """
        Render a comment body, returning the rendered html.

        @param value: comment text to render
        @type value: L{str}
        @param to_root: same as in the templates
        @type to_root: L{str}
        @return: the rendered HTML of the comment text
        @rtype: L{str}
        """
        rewritten = self.filemanager.rewrite_urls_in_text(
            value,
            to_root=to_root,
        )
        rewritten = self.reference_rewriter.rewrite_urls_in_text(
            rewritten,
            to_root=to_root,
        )
        rendered = self.markdown(
            rewritten,
        )
        return rendered

    def _render_license_text_filter(self, value):
        """
        Render a license text, returning the rendered html.

        @param value: license text to render
        @type value: L{str}
        @return: the rendered HTML of the license text
        @rtype: L{str}
        """
        rendered = mistune.html(value)
        return rendered

    def _format_date(self, value):
        """
        Format a date.

        @param value: date to format
        @type value: L{datetime.datetime}
        @return: the formated date
        @rtype: L{str}
        """
        return value.strftime("%Y-%m-%d")

    def _format_timestamp(self, value, allow_none=False):
        """
        Format a timestamp to a human readbale date.

        @param value: timestamp to format
        @type value: L{int}
        @param allow_none: if nonzero, allow passing None as a value
        @type allow_none: L{bool}
        @return: the formated date and time
        @rtype: L{str}
        """
        if value is None:
            if allow_none:
                return "-"
            else:
                raise TypeError("Got None as value with allow_none=False!")
        return datetime.datetime.fromtimestamp(value).isoformat()

    def _first_elements(self, value, n):
        """
        Return the first n elements in a value

        @param value: list whose first elements should be returned
        @type value: L{list} or L{tuple}
        @param n: number of elements to return
        @type n: L{int}
        @return: the first n elements in value
        @rtype: L{list}
        """
        return list(value)[:n]

    def _default_index(self, value, i, default):
        """
        Return value[i] if value is defined and value[i] exists, otherwise default.

        @param value: element to get i-th element of
        @type value: any indexable or L{jinja2.Undefined}
        @param i: index/key to get
        @type i: any
        @param default: default value to return
        @type default: any
        @return: value[i] if value is defined and value[i] exists, otherwise default
        @rtype: type of value[i] or default
        """
        if isinstance(value, Undefined):
            return default
        try:
            return value[i]
        except (KeyError, IndexError):
            return default

    def _first_nonzero_filter(self, value):
        """
        Returns the first nonzero element in a list or the last one if none of them are nonzero.

        @param value: value to process
        @type value: L{list}
        @return: the first nonzero element in the list or the last element
        @rtype: as value[x]
        """
        for v in value:
            if v:
                return v
        return v

    def _stickysort_filter(self, comments):
        """
        Sort a list of comments such that stickied comments come first.

        @param comments: comments to sort
        @type comments: L{list} of L{arcticzim.db.models.Comment}
        @return: the sorted comments
        @type comments: L{list} of L{arcticzim.db.models.Comment}
        """
        return sorted(comments, reverse=True, key=lambda x: (x is True))

    def _render_postsummary_by_url(self, url, to_root):
        """
        Render a postsummary for a post with a specific url.

        @param url: (reddit) url to post
        @type url: L{str}
        @param to_root: same as in templates
        @type to_root: L{str}
        @return: the rendered post summary
        @rtype: L{str}
        """
        reference = parse_reddit_url(url)
        if (reference is None) or (reference["type"] != "post"):
            raise ValueError("Error: not a reddit reference")
        post_id = reference["post"]
        return self.worker.directly_render_postsummary(post_id, to_root=to_root)

    # =========== tests ===============

    def _is_date(self, obj):
        """
        Return True if obj is a date or datetime.

        @param obj: object to check
        @type obj: any
        @return: True if obj is a date or datetime
        @rtype: L{bool}
        """
        return isinstance(obj, (datetime.date, datetime.datetime))

    def _is_local_post_url(self, url):
        """
        Return True if url is a url referencing an post that is locally available.

        @param url: url to check
        @type url: L{str}
        @return: whether the url points to a post that is locally available
        @rtype: L{bool}
        """
        if not isinstance(url, str):
            return False
        parsed = parse_reddit_url(url)
        if parsed is None:
            return False
        if parsed["type"] != "post":
            return False
        return self.reference_rewriter.should_rewrite(parsed)



"""
This module contains the code for downloading media files.
"""
import os
import hashlib
import time
import json
from mimetypes import guess_type
from urllib.parse import urlparse, parse_qsl, unquote_plus, urlunparse, urlencode

from sqlalchemy import select, func
from sqlalchemy.orm import undefer, selectinload
import requests
import tqdm
from yt_dlp import YoutubeDL
from redvid import Downloader as RedvidDL

from .db.models import MediaFile, Post, Comment
from .imgutils import minimize_image, mimetype_is_image, mimetype_is_video
from .util import get_urls_from_string


class DownloadFailed(Exception):
    """
    Exception raised when a download failed.
    """
    pass


def unify_url(url):
    """
    Transform an URL into a "unified" URL used to identify similiar URLs.

    The resulting URL may not be valid and is only supposed to be used
    for identifying if two URLs are the same.

    @param url: URL to unify
    @type url: L{str}
    @return: a unified URL
    @rtype: L{str}
    """
    # modified version of https://stackoverflow.com/a/9468284
    if isinstance(url, bytes):
        url = url.decode("utf-8")
    if url == "":
        return ""
    parts = urlparse(url)
    _query = urlencode(
        list(sorted(frozenset(parse_qsl(parts.query)))),
    )
    _path = parts.path
    while True:
        # using a loop here as a URL may be urlencoded multiple times
        # and we need the output here to be stable
        unquoted = unquote_plus(_path)
        if unquoted == _path:
            break
        _path = unquoted
    parts = parts._replace(
        query=_query,
        path=_path,
        fragment="",
        scheme="http",
    )
    return urlunparse(parts)


def hash_url(url):
    """
    Hash a url, returning the hexdigest.

    @param url: URL to hash
    @type url: L{str}
    @return: the hashed url
    @rtype: L{str}
    """
    hasher = hashlib.md5(unify_url(url).encode("utf-8"))
    return hasher.hexdigest()


def has_downloaded(session, url, any_status=True):
    """
    Check if the URL has already been downloaded.

    @param session: sqlalchemy session
    @type session: L{sqlalchemy.orm.Session}
    @param url: url to check
    @type url: L{str}
    @param any_status: if nonzero, return True as long as the download has been attempted
    @type any_status: L{bool}
    @return: whether the file has been downloaded or not
    @rtype: L{bool}
    """
    unified_url = unify_url(url)
    mf = session.execute(
        select(MediaFile).where(
            (MediaFile.url == unified_url),
            ((MediaFile.downloaded == True) if not any_status else True),
        )
    ).one_or_none()
    return (mf is not None)


def download(session, url, mediadir, enable_post_processing=True, download_videos=True, max_image_dimension=512):
    """
    Download the media at the specified URL.

    @param session: sqlalchemy session
    @type session: L{sqlalchemy.orm.Session}
    @param url: url to download
    @type url: L{str}
    @param mediadir: directory where files should be downloaded too
    @type mediadir: L{str}
    @param enable_post_processing: whether post-processing should be applied
    @type enable_post_processing: L{bool}
    @param download_videos: whether videos should be downloaded
    @type download_videos: L{bool}
    @param max_image_dimension: how many pixels the wider side of an image may have at most
    @type max_image_dimension: L{int}
    """
    url_hash = hash_url(url)
    outpath = os.path.join(mediadir, url_hash)
    guessed_mimetype = guess_type(urlparse(url).path)[0]
    is_probably_image = (guessed_mimetype is not None) and (guessed_mimetype.startswith("image/"))
    try:
        if (is_ytdlp(url) or is_redvid(url)) and not is_probably_image:
            if not download_videos:
                return
            if is_redvid(url):
                mimetype = do_redvid_download(url=url, mediadir=mediadir, outpath=outpath)
            else:
                mimetype = do_ytldp_download(url=url, mediadir=mediadir, outpath=outpath)
            hasher = hashlib.md5()
            size = 0
            with open(outpath, "rb") as fin:
                chunk = True
                while chunk:
                    chunk = fin.read(4096)
                    size += len(chunk)
                    hasher.update(chunk)
            md5 = hasher.hexdigest()
        else:
            headers = {
                "user-agent": "Mozilla/5.0 (X11; Linux x86_64; rv:140.0) Gecko/20100101 Firefox/140.0",
            }
            try:
                r = requests.get(url, headers=headers, stream=True)
                r.raise_for_status()
            except Exception as e:
                raise DownloadFailed() from e
            mimetype = r.headers.get("content-type", None)
            if mimetype is None:
                mimetype = guess_type(urlparse(url).path)[0]
            elif ";" in mimetype:
                mimetype = mimetype[:mimetype.find(";")]
            hasher = hashlib.md5()
            size = 0
            with open(outpath, "wb") as fout:
                for chunk in r.iter_content(chunk_size=4096):
                    fout.write(chunk)
                    hasher.update(chunk)
                    size += len(chunk)
            md5 = hasher.hexdigest()
    except DownloadFailed:
        mo = MediaFile(
            url=unify_url(url),
            downloaded=False,
        )
        session.add(mo)
        session.commit()
        return

    existing_mf = session.execute(
        select(MediaFile).where(
            MediaFile.md5 == md5,
            MediaFile.downloaded == True,
            MediaFile.primary_uid is None,
        )
    ).one_or_none()
    if existing_mf is not None:
        # file already downloaded
        os.remove(os.path.join(mediadir, url_hash))
        mf = MediaFile(
            url=unify_url(url),
            downloaded=True,
            md5=md5,
            mimetype=mimetype,
            primary_uid=existing_mf[0].uid,
        )
        session.add(mf)
        session.commit()
        return
    else:
        mf = MediaFile(
            url=unify_url(url),
            downloaded=True,
            md5=md5,
            mimetype=mimetype,
            size=size,
        )
        if enable_post_processing:
            post_process(mediadir, mf, max_image_dimension=max_image_dimension)
        session.add(mf)
        session.commit()


def is_ytdlp(url):
    """
    Check if the target url should be downloaded via yt-dlp.

    @param url: url to check
    @type url: L{bool}
    @return: whether yt-dlp should be used to download target url
    @rtype: L{bool}
    """
    with YoutubeDL(params={"allowed_extractors": ["default", "-generic"]}) as yt:
        for ie in yt._ies.values():  # TODO: private attribute access is probably a bad idea
            if ie.suitable(url) and ie.working():
                return True
    return False


def do_ytldp_download(url, mediadir, outpath):
    """
    Download a URL with yt-dlp.

    @param url: url to download
    @type url: L{str}
    @param mediadir: directory where files should be downloaded too
    @type mediadir: L{str}
    @param outpath: path the file should be written to
    @type outpath: L{str}
    @return: the mimetype of the downloaded video
    @rtype: L{str}
    """
    hashed_url = hash_url(url)
    params = {
        # TODO: change "worst" to smallest filesize
        "format": "worst/worstvideo+worstaudio",
        "outtmpl": "{}.%(ext)s".format(outpath),
    }
    with YoutubeDL(params=params) as yt:
        try:
            yt.download(url)
        except Exception as e:
            raise DownloadFailed("yt-dlp raised an exception when attempting to download video") from e
    # todo: the following is probably inefficient for large file collections
    for fname in os.listdir(mediadir):
        if fname.startswith(hashed_url):
            break
    mimetype = guess_type(fname)[0] or "video/mp4"
    os.rename(os.path.join(mediadir, fname), outpath)
    return mimetype


def is_redvid(url):
    """
    Check if the target url should be downloaded via redvid.

    @param url: url to check
    @type url: L{bool}
    @return: whether redvid should be used to download target url
    @rtype: L{bool}
    """
    return ("://v.redd.it" in url)


def do_redvid_download(url, mediadir, outpath):
    """
    Download a URL with redvid.

    @param url: url to download
    @type url: L{str}
    @param mediadir: directory where files should be downloaded too
    @type mediadir: L{str}
    @param outpath: path the file should be written to
    @type outpath: L{str}
    @return: the mimetype of the downloaded video
    @rtype: L{str}
    """
    downloader = RedvidDL(url=url, path=mediadir, min_q=True)
    try:
        path = downloader.download()
    except Exception as e:
        raise DownloadFailed("redvid raised an exception when attempting to download video") from e
    mimetype = guess_type(path)[0] or "video/mp4"
    os.rename(path, outpath)
    return mimetype


def post_process(mediadir, mediafile, max_image_dimension=512):
    """
    Post process an image file.

    Don't forget to add and commit the mediafile afterwards!

    @param mediadir: directory containing the files
    @type mediadir: L{str}
    @param mediafile: the mediafile object
    @type mediafile: L{arcticzim.db.models.MediaFile}
    @param max_image_dimension: how many pixels the wider side of an image may have at most
    @type max_image_dimension: L{int}
    """
    path = os.path.join(mediadir, hash_url(mediafile.url))
    if mimetype_is_image(mediafile.mimetype):
        mimetype, size = minimize_image(
            path,
            max_w=max_image_dimension,
            max_h=max_image_dimension,
        )
        mediafile.size = size
        mediafile.mimetype = mimetype


def get_urls_from_post(post, include_reddit_videos=True, include_external_videos=False, include_comments=False):
    """
    Return all URLs used in a post

    @param post: post to get URLs from
    @type post: L{arcticzim.db.models.Post}
    @param include_reddit_videos: whether reddit videos should be included
    @type include_reddit_videos: L{bool}
    @param include_external_videos: whether non-reddit videos should be included
    @type include_external_videos: L{bool}
    @param include_comments: if nonzero, download media in comments too
    @type include_comments: L{bool}
    @return: a list of posts
    @rtype: L{list} of L{str}
    """
    urls = []
    if post.post_hint in ("rich:video", ) and include_external_videos:
        urls.append(post.url)
    if post.post_hint in ("hosted:video", ) and include_reddit_videos:
        urls.append(post.url)
    if post.post_hint in ("image", ):
        urls.append(post.url)
    if post.selftext:
        for url in get_media_urls_from_string(
            post.selftext,
            include_reddit_videos=include_reddit_videos,
            include_external_videos=include_external_videos,
        ):
            urls.append(url)
    if post.is_gallery:
        media_metadata = json.loads(post.media_metadata)
        if media_metadata:
            for img_data in media_metadata.values():
                if "s" in img_data:
                    urls.append(img_data["s"]["u"])
                elif "p" in img_data:
                    urls.append(img_data["p"][-1]["u"])
    if include_comments:
        for comment in post.comments:
            urls += get_media_urls_from_string(comment.body)
    return urls


def get_media_urls_from_string(s, include_reddit_videos=True, include_external_videos=False):
    """
    Find all media URLs matching certain conditions in a string and return them

    @param s: string to search for URLs
    @type s: L{str}
    @param include_reddit_videos: whether reddit videos should be included
    @type include_reddit_videos: L{bool}
    @param include_external_videos: whether non-reddit videos should be included
    @type include_external_videos: L{bool}
    @return: a list of URLs found
    @rtype: L{list} of L{str}
    """
    urls = get_urls_from_string(s)
    found_urls = []
    for url in urls:
        mimetype = guess_type(urlparse(url).path)[0]
        if mimetype is None:
            # no guess
            continue
        if mimetype.startswith("video/"):
            if is_redvid(url):
                if include_reddit_videos:
                    found_urls.append(url)
            elif is_ytdlp(url):
                if include_external_videos:
                    found_urls.append(url)
        elif mimetype.startswith("image/"):
            found_urls.append(url)
    return found_urls


def download_all(
    session,
    mediadir,
    sleep=0.5,
    enable_post_processing=True,
    download_reddit_videos=True,
    download_external_videos=False,
    include_comments=False,
    max_image_dimension=512,
):
    """
    Download all files of posts.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param mediadir: directory to store media in
    @type mediadir: L{str}
    @param sleep: sleep time between downloads in seconds
    @type sleep: L{int} or L{float}
    @param download_reddit_videos: whether reddit videos should be downloaded
    @type download_reddit_videos: L{bool}
    @param download_external_videos: whether non-reddit videos should be downloaded
    @type download_external_videos: L{bool}
    @param enable_post_processing: whether post-processing should be applied
    @type enable_post_processing: L{bool}
    @param include_comments: if nonzero, download media in comments too
    @type include_comments: L{bool}
    @param max_image_dimension: how many pixels the wider side of an image may have at most
    @type max_image_dimension: L{int}
    """
    n = session.execute(select(func.count(Post.uid))).one()[0]
    stmt = select(Post).options(
        undefer(Post.url),
        undefer(Post.selftext),
        selectinload(Post.comments),
        undefer(Post.comments, Comment.body),
    ).execution_options(
        yield_per=1000,
    )
    for post in tqdm.tqdm(session.execute(stmt).scalars(), desc="Searching posts", total=n, unit="posts"):
        urls = get_urls_from_post(
            post,
            include_reddit_videos=download_reddit_videos,
            include_external_videos=download_external_videos,
            include_comments=include_comments,
        )
        urls = set([url for url in urls if not has_downloaded(session=session, url=url)])
        if not urls:
            continue
        for url in tqdm.tqdm(urls, desc="downloading files for {}".format(post.id), total=len(urls), unit="files"):
            download(
                session=session,
                url=url,
                mediadir=mediadir,
                enable_post_processing=enable_post_processing,
                download_videos=(download_external_videos or download_reddit_videos),
                max_image_dimension=max_image_dimension,
            )
            time.sleep(sleep)


class MediaFileManager(object):
    """
    This class managers the conversion from external to internal links.

    @ivar session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @ivar enabled: if the rewrite is enabled in the first place
    @type enabled: L{bool}
    @ivar images_enabled: if media links should be rewritten
    @type images_enabled: L{bool}
    @ivar videos_enabled: if video links should be rewritten
    @type videos_enabled: L{bool}
    @ivar referenced_files: a list of files referenced by the rewritten URLs
    @type referenced_files: L{list} of L{int}
    """
    def __init__(self, session, enabled=True, images_enabled=True, videos_enabled=True):
        """
        The default constructor.

        @param session: sqlalchemy session to use
        @type session: L{sqlalchemy.orm.Session}
        @param enabled: if the rewrite is enabled in the first place
        @type enabled: L{bool}
        @param images_enabled: if media links should be rewritten
        @type images_enabled: L{bool}
        @param videos_enabled: if video links should be rewritten
        @type videos_enabled: L{bool}
        """
        self.session = session
        self.enabled = enabled
        self.images_enabled = images_enabled
        self.videos_enabled = videos_enabled
        self.referenced_files = []

    def reset(self):
        """
        Reset internal states.
        """
        self.referenced_files = []

    def should_rewrite(self, mediafile):
        """
        Check whether an URL should be rewritten.

        This method does not check if the local file exists. That is
        presumed to have already been done. Instead, this method checks
        whether we want to include the file.

        @param mediafile: mediafile to check
        @type mediafile: L{arcticzim.db.models.MediaFile}
        @return: whether the link should be rewritten
        @rtype: L{str}
        """
        if not self.enabled:
            return False
        if mimetype_is_image(mediafile.mimetype) and self.images_enabled:
            return True
        if mimetype_is_video(mediafile.mimetype) and self.videos_enabled:
            return True
        return False

    def rewrite_url(self, url, to_root):
        """
        Rewrites an external URL to an internal one if the file exists locally.

        @param url: url to rewrite
        @type url: L{str}
        @param to_root: like with the renderer, a prefix (e.g. C{../..}) that indicates how to navigate to the root directory
        @type to_root: L{str}
        @return: the url to use, which may still be an external one
        @rtype: L{str}
        """
        unified_url = unify_url(url)
        if not unified_url:
            # we should not rewrite empty URLs
            return url
        mf = self.session.execute(select(MediaFile).where(MediaFile.url == unified_url)).one_or_none()
        if mf is None:
            # file not downloaded, can't rewrite
            return url
        mf = mf[0]
        if mf.primary_uid is not None:
            mf = mf.primary
        if not mf.downloaded:
            # file not downloaded, can't rewrite
            return url
        if not self.should_rewrite(mf):
            # we should not use this media file
            return url
        new_url = "{}/media/{}".format(to_root, mf.uid)
        self.referenced_files.append(mf.uid)
        return new_url

    def rewrite_urls_in_text(self, text, to_root):
        """
        Rewrite all external URLs found in the text with internal ones if the files exist locally.

        @param text: text to rewrite
        @type text: L{str}
        @param to_root: like with the renderer, a prefix (e.g. C{../..}) that indicates how to navigate to the root directory
        @type to_root: L{str}
        @return: the rewritten text
        @rtype: L{str}
        """
        urls = get_media_urls_from_string(text)
        for url in urls:
            new_url = self.rewrite_url(url, to_root=to_root)
            if new_url != url:
                text = text.replace(url, new_url)
        return text



"""
This module contains functionality to create batches from a stream of data.
"""


class BucketMaker(object):
    """
    The BucketMaker takes single elements and creates new buckets from them.

    @ivar maxsize: max size of a bucket
    @type maxsize: L{int}
    """
    def __init__(self, maxsize):
        """
        The default constructor.

        @param maxsize: max size of a bucket
        @type maxsize: L{int}
        """
        assert isinstance(maxsize, int) and maxsize > 0
        self.maxsize = maxsize
        self._cur_bucket = []

    def feed(self, element):
        """
        Add a new element to the current bucket.

        If the current bucket will be full as part of the task, return
        the bucket and start a new one. Otherwise, return None.


        @param element: element to put into a bucket
        @type element: any
        @return: the bucket if the bucket is now full else L{None}
        @rtype: L{list} if bucket is full else L{None}
        """
        self._cur_bucket.append(element)
        if len(self._cur_bucket) >= self.maxsize:
            # new bucket
            bucket = self._cur_bucket
            self._cur_bucket = []
            return bucket
        return None

    def finish(self):
        """
        If the current bucket is not empty, return it.

        @return: the current bucket if it is not empty else L{None}
        @rtype: L{list} or L{None}
        """
        if self._cur_bucket:
            bucket = self._cur_bucket
            self._cur_bucket = []
            return bucket
        return None



"""
Script for generating the icons.
"""
from PIL import Image, ImageDraw, ImageFont

ICONS = [
    # tuples of (name, filename, text)
    # if a filename is specified, it should be an RGBA image where the foreground is #dddddd and the background transparent
    ("img", "camera-4-512.png", None),
    ("video", "play-512.png", None),
    ("text", None, "T"),
    ("poll", "pie-chart-512.png", None),
    ("link", None, "L"),
]

ICON_COLOR = "darkgray"
ICON_SIZE = (512, 512)
ICON_PASTE_SIZE = (400, 400)
FONT_SIZE = 400

for icon in ICONS:
    img = Image.new(mode="RGBA", size=ICON_SIZE, color=(0, 0, 0, 0))
    draw = ImageDraw.Draw(img)
    draw.rounded_rectangle(
        xy=((0, 0), ICON_SIZE),
        radius=(ICON_SIZE[0] // 6),
        fill=ICON_COLOR,
    )
    if icon[1] is not None:
        with Image.open(icon[1]) as other_img:
            resized = other_img.resize(ICON_PASTE_SIZE)
            img.alpha_composite(
                resized,
                (
                    (ICON_SIZE[0] - ICON_PASTE_SIZE[0]) // 2,
                    (ICON_SIZE[1] - ICON_PASTE_SIZE[1]) // 2,
                ),
            )
    if icon[2] is not None:
        font = ImageFont.truetype("LiberationSerif-Bold.ttf", FONT_SIZE)
        draw.text(
            xy=(ICON_SIZE[0] // 2, ICON_SIZE[1] // 2),
            text=icon[2],
            fill=(255, 255, 255, 255),
            anchor="mm",
            font=font,
        )
    img.save("icon_{}.png".format(icon[0]))



"""
The worker logic for multi process rendering.

Workers receive their tasks from an inqueue fed from the builder. They
process the tasks by loading the required objects and feed them to a
renderer. The result is put into the outqueue, where the builder will
take the results and add them to the creator.

@var MARKER_WORKER_STOPPED: a symbolic constant put into the output queue when the worker is finished
@type MARKER_WORKER_STOPPED: L{str}
@var MARKER_TASK_COMPLETED: a symbolic constant put into the output queue when a task was completed
@type MARKER_TASK_COMPLETED: L{str}

@var MAX_POST_EAGERLOAD: when loading subreddits, do not eagerload if more than this number of posts are in said object
@type MAX_POST_EAGERLOAD: L{int}
@var POST_LIST_YIELD: number of story to fetch at once when rendering posts, ...
@type POST_LIST_YIELD: L{int}
"""
import contextlib
import os
import time

from sqlalchemy import select, func, desc
from sqlalchemy.orm import Session, undefer
from sqlalchemy.orm import joinedload, selectinload, noload, raiseload

try:
    import memray
except ImportError:
    memray = None

from .renderer import HtmlRenderer, RenderResult, SubredditInfo, SUBREDDITS_ON_INDEX_PAGE
from .statistics import query_post_stats
from ..util import ensure_iterable
from ..downloader import MediaFileManager
from ..fetcher import ReferenceUrlRewriter
from ..db.models import Post, User, Comment, Subreddit, WikiPage, SubredditRule


MARKER_WORKER_STOPPED = "stopped"
MARKER_TASK_COMPLETED = "completed"

MAX_POST_EAGERLOAD = 10000
MIN_POSTS_FOR_EXPLICIT_STATS = 10000
MIN_POSTS_FOR_STREAM = 10000

POST_LIST_YIELD = 2000


class Task(object):
    """
    Base class for all worker tasks.

    @cvar type: type of task
    @type type: L{str}
    """
    type = "<unset>"

    @property
    def name(self):
        """
        Return a non-unqiue id describing this task.

        @return: a name describing this task
        @rtype: L{str}
        """
        return "<unknown>"


class StopTask(Task):
    """
    Task indicating that the worker should shut down.
    """
    type = "stop"

    @property
    def name(self):
        return "stop"


class PostRenderTask(Task):
    """
    A L{Task} for rendering posts.

    @ivar post_uids: ids of posts to render, as list of post uids
    @type post_uids: L{list} of L{int}
    """
    type = "post"

    def __init__(self, post_uids):
        """
        The default constructor.

        @param post_uids: ids of stories to render, as list of post uids
        @type post_uids: L{list} of L{int}
        """
        assert isinstance(post_uids, (list, tuple))
        assert isinstance(post_uids[0], int)
        self.post_uids = post_uids

    @property
    def name(self):
        if len(self.post_uids) == 0:
            return "post_empty"
        return "posts_{}+{}".format(
            self.post_uids[0],
            len(self.post_uids) - 1,
        )


class SubredditRenderTask(Task):
    """
    A L{Task} for rendering a subreddit.

    @ivar subreddit_name: name of subreddit to render
    @type subreddit_name: L{str}
    @ivar subtask: subtask (e.g. sort order) to render
    @type subtask: L{str}
    """
    type = "subreddit"

    def __init__(self, subreddit_name, subtask):
        """
        The default constructor.

        @param subreddit_name: name of subreddit to render
        @type subreddit_name: L{str}
        @param subtask: subtask (e.g. sort order, wiki, stats) to render
        @type subtask: L{str}
        """
        assert isinstance(subreddit_name, str)
        assert isinstance(subtask, str)
        self.subreddit_name = subreddit_name
        self.subtask = subtask

    @property
    def name(self):
        return "{}_r/{}".format(
            self.subtask,
            self.subreddit_name
        )


class UserRenderTask(Task):
    """
    A L{Task} for rendering a user.

    @ivar username: username of user to render
    @type username: L{str}
    @ivar part: part (post, comments, stats) to render
    @type part: L{str}
    @ivar sort: sort order to render
    @type sort: L{str}
    """
    type = "user"

    def __init__(self, username, part, sort):
        """
        The default constructor.

        @param username: username of user to render
        @type username: L{int}
        @param part: part (post, comments, stats) to render
        @type part: L{str}
        @param sort: sort order to render
        @type sort: L{str}
        """
        assert isinstance(username, str)
        assert isinstance(part, str)
        assert isinstance(sort, str) or (sort is None)
        self.username = username
        self.part = part
        self.sort = sort

    @property
    def name(self):
        return "{}_{}_{}".format(
            self.username,
            self.part,
            self.sort,
        )


class EtcRenderTask(Task):
    """
    A L{Task} for rendering specific, individual pages.

    @ivar subtask: the name of the subtask to perform (e.g. index)
    @type subtask: L{str}
    """
    type = "etc"

    def __init__(self, subtask):
        """
        The default constructor.

        @param subtask: name of the subtask to perform
        @type subtask: L{str}
        """
        assert isinstance(subtask, str)
        self.subtask = subtask

    @property
    def name(self):
        return "etc_{}".format(self.subtask)


class WorkerOptions(object):
    """
    Options for the worker.

    @ivar eager: eager load objects from database
    @type eager: L{bool}
    @ivar log_directory: if not None, enable logging and write log here
    @type log_directory: L{str} or L{None}
    @ivar memprofile_directory: if not None, profile memory usage and write files into this directory
    @type memprofile_directory: L{str} or L{None}
    @ivar with_stats: if nonzero, include statistics
    @type with_stats: L{bool}
    @ivar with_media: if nonzero, include media
    @type with_media: L{bool}
    @ivar with_videos: if nonzero, include videos
    @type with_videos: L{bool}
    """
    def __init__(
        self,
        eager=True,
        log_directory=None,
        memprofile_directory=None,

        with_stats=True,
        with_media=True,
        with_videos=True,
    ):
        """
        The default constructor.

        @param eager: if nonzero, eager load objects from database
        @type eager: L{bool}
        @param log_directory: if specified, enable logging and write log here
        @type log_directory: L{str} or L{None}
        @param memprofile_directory: if specified, profile memory usage and write files into this directory
        @type memprofile_directory: L{str} or L{None}
        @param with_stats: if nonzero, include statistics
        @type with_stats: L{bool}
        @param with_media: if nonzero, include media
        @type with_media: L{bool}
        @param with_videos: if nonzero, include videos
        @type with_videos: L{bool}
        """
        assert isinstance(log_directory, str) or (log_directory is None)
        assert isinstance(memprofile_directory, str) or (memprofile_directory is None)
        self.eager = eager
        self.log_directory = log_directory
        self.memprofile_directory = memprofile_directory
        self.with_stats = with_stats
        self.with_media = with_media
        self.with_videos = with_videos


class Worker(object):
    """
    The worker should be instantiated in a new process, where it will
    continuously process tasks from the main process.

    @ivar id: id of this worker
    @type id: L{int}
    @ivar inqueue: the queue providing new tasks
    @type inqueue: L{multiprocessing.Queue}
    @ivar outqueue: the queue where results will be put
    @type outqueue: L{multiprocessing.Queue}
    @ivar renderer: the renderer used to render the content
    @type renderer: L{arcticzim.zimbuild.renderer.HtmlRenderer}
    @ivar engine: engine used for database connection
    @type engine: L{sqlalchemy.engine.Engine}
    @ivar session: database session
    @type session: L{sqlalchemy.orm.Session}
    @ivar options: options for this worker
    @type options: L{WorkerOptions}

    @ivar _log_file: file used for logging
    @type _log_file: file-like object
    @ivar _last_log_time: timestamp of last log entry
    @ivar _last_log_time: L{int}
    """
    def __init__(self, id, inqueue, outqueue, engine, options, render_options):
        """
        The default constructor.

        @param id: id of this worker
        @type id: L{int}
        @param inqueue: the queue providing new tasks
        @type inqueue: L{multiprocessing.Queue}
        @param outqueue: the queue where results will be put
        @type outqueue: L{multiprocessing.Queue}
        @param engine: engine used for database connection. Be sure to dispose the connection pool first!
        @type engine: L{sqlalchemy.engine.Engine}
        @param options: options for the worrker
        @type options: L{WorkerOptions}
        @param render_options: options for the renderer
        @type render_options: L{arcticzim.zimbuild.renderer.RenderOptions}
        """
        assert isinstance(id, int)
        self.id = id
        self.inqueue = inqueue
        self.outqueue = outqueue
        self.engine = engine

        self.session = Session(engine)
        self.options = options
        self.renderer = HtmlRenderer(
            worker=self,
            options=render_options,
            filemanager=MediaFileManager(
                session=self.session,
                enabled=self.options.with_media,
                images_enabled=self.options.with_media,
                videos_enabled=self.options.with_videos,
            ),
            reference_rewriter=ReferenceUrlRewriter(
                session=self.session,
            ),
        )

        self.setup_logging()

        self.log("Worker initialized.")

    def setup_logging(self):
        """
        Setup the logging system.
        """
        self._last_log_time = time.time()
        if self.options.log_directory is not None:
            fn = os.path.join(
                self.options.log_directory,
                "log_worker_{}.txt".format(self.id),
            )
            self._log_file = open(fn, mode="w", encoding="utf-8")
        else:
            self._log_file = None

    def log(self, msg):
        """
        Log a message.

        @param msg: message to log
        @type msg: L{str}
        """
        assert isinstance(msg, str)
        if self._log_file is not None:
            full_msg = "[{}][+{:8.3f}s] {}\n".format(
                time.ctime(),
                time.time() - self._last_log_time,
                msg,
            )
            self._log_file.write(full_msg)
            self._log_file.flush()
        self._last_log_time = time.time()

    def run(self):
        """
        Run the worker.

        This does not start a new process, the worker should already
        have been instantiated in a new process.

        This method will run in a loop, taking new tasks from the inqueue,
        processing them and putting the results in the outqueue until a
        L{StopTask} has been received. Upon completion,
        L{MARKER_WORKER_STOPPED} will be put on the outqueue once.
        Additionally, L{MARKER_TASK_COMPLETED} is put in the queue
        after each task.
        """
        self.log("Entering mainloop.")
        running = True
        while running:
            task = self.inqueue.get(block=True)
            self.log("Received task '{}'".format(task.name))
            with self.get_task_processing_context(task=task):
                if task.type == "stop":
                    # stop the worker
                    self.log("Stopping worker...")
                    running = False
                    self._cleanup()
                elif task.type == "post":
                    self.process_post_task(task)
                elif task.type == "user":
                    self.process_user_task(task)
                elif task.type == "subreddit":
                    self.process_subreddit_task(task)
                elif task.type == "etc":
                    self.process_etc_task(task)
                else:
                    raise ValueError("Task {} has an unknown task type '{}'!".format(repr(task), task.type))
                if task.type != "stop":
                    # notify builder that a task was completed
                    self.log("Marking task as completed.")
                    self.outqueue.put(MARKER_TASK_COMPLETED)
        # send a message indicated that this worker has finished
        self.outqueue.put(MARKER_WORKER_STOPPED)
        self.log("Worker finished.")

    def _cleanup(self):
        """
        Called before the worker finishes.

        All cleanup (e.g. closing sessions) should happen here.
        """
        self.session.close()
        self.engine.dispose()

    def handle_result(self, result):
        """
        Handle a renderer result, putting it in the outqueue.

        @param result: result to handle
        @type result: L{arcticzim.zimbuild.renderer.RenderResult} or iterable of it
        """
        it = ensure_iterable(result)
        for i, subresult in enumerate(it):
            self.log("Submitting result part {}...".format(i))
            self.outqueue.put(subresult)

    def get_task_processing_context(self, task):
        """
        Return A context manager that runs while a task is being processed.

        @param task: task the context is for
        @type task: Task
        @return: context manager to use
        @rtype: a context manager
        """
        if self.options.memprofile_directory is not None:
            if memray is None:
                raise ImportError("Could not import package 'memray' required for memory profiling!")
            file_name = os.path.join(
                self.options.memprofile_directory,
                "mp_{}_{}.bin".format(self.id, task.name)
            )
            return memray.Tracker(
                destination=memray.FileDestination(
                    path=file_name,
                    overwrite=True,
                    compress_on_exit=False,
                ),
                native_traces=False,
                trace_python_allocators=False,
                follow_fork=False,  # already in fork
            )
        else:
            return contextlib.nullcontext()

    def process_post_task(self, task):
        """
        Process a received post task.

        @param task: task to process
        @type task: L{PostRenderTask}
        """
        # setup load options
        if self.options.eager:
            options = (
                undefer(Post.selftext),
                undefer(Post.title),
                selectinload(Post.comments).undefer(Comment.body),
                joinedload(Post.subreddit),
                joinedload(Post.author),
            )
        else:
            options = (
                undefer(Post.selftext),
                undefer(Post.title),
                selectinload(Post.comments).undefer(Comment.body),
            )
        for post_uid in task.post_uids:
            # get the post
            self.log("Retrieving post...")
            post = self.session.scalars(
                select(Post)
                .where(Post.uid == post_uid)
                .options(
                    *options,
                )
            ).first()
            self.log("Retrieved post, rendering...")
            result = self.renderer.render_post(post)
            self.log("Rendered post, submitting result...")
            self.handle_result(result)
            self.log("Done.")

    def directly_render_postsummary(self, post_id, to_root):
        """
        Render a post summary, directly returning the HTML.

        This is a helper function for rendering crossposts. It is called
        by the renderer to load the post and will call the renderer again.

        @param post_id: id of post to process
        @type post_id: L{str}
        @param to_root: rootification prefix (see templates)
        @type to_root: L{str}
        @return: the rendered html
        @rtype: L{str}
        """
        # setup load options
        if self.options.eager:
            options = (
                undefer(Post.selftext),
                undefer(Post.title),
                selectinload(Post.comments).undefer(Comment.body),
                joinedload(Post.subreddit),
                joinedload(Post.author),
            )
        else:
            options = (
                undefer(Post.selftext),
                undefer(Post.title),
                selectinload(Post.comments).undefer(Comment.body),
            )
        self.log("Retrieving post for postsummary...")
        post = self.session.scalars(
            select(Post)
            .where(Post.id == post_id)
            .options(
                *options,
            )
        ).first()
        self.log("Retrieved post, rendering postsummary...")
        result = self.renderer.directly_render_post_summary(post, to_root=to_root)
        return result

    def process_subreddit_task(self, task):
        """
        Process a received subreddit task.

        @param task: task to process
        @type task: L{SubredditRenderTask}
        """
        if task.subtask == "stats":
            self.process_subreddit_stats_task(task)
        elif task.subtask == "wiki":
            self.process_subreddit_wiki_task(task)
        elif task.subtask == "rules":
            self.process_subreddit_rules_task(task)
        else:
            self.process_subreddit_sort_task(task)

    def process_subreddit_sort_task(self, task):
        """
        Process a received subreddit sort task.

        Note: this is called from L{Worker.process_subreddit_task}.

        @param task: task to process
        @type task: L{SubredditRenderTask}
        """
        # count posts in subreddit
        self.log("Counting posts in subreddit...")
        count_stmt = (
            select(func.count(Post.uid))
            .where(
                Post.subreddit_name == task.subreddit_name,
            )
        )
        n_posts_in_subreddit = self.session.execute(count_stmt).scalar_one()
        self.log("Found {} posts.".format(n_posts_in_subreddit))
        # load subreddit
        self.log("Loading subreddit...")
        subreddit_stmt = (
            select(Subreddit)
            .where(
                Subreddit.name == task.subreddit_name,
            )
            .options(
                raiseload(Subreddit.posts),
                raiseload(Subreddit.comments),
                joinedload(Subreddit.wikipages),
            )
        )
        subreddit = self.session.scalars(subreddit_stmt).first()
        self.session.expunge(subreddit)  # prevent subreddit from being modified and storing objects
        self.log("Subreddit loaded.")
        if subreddit is None:
            self.log("-> Subreddit not found!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return

        # load posts
        self.log("Starting to load posts...")
        # always use eager loading, lazy is horrible for performance here
        options = (
            selectinload(Post.comments),
            joinedload(Post.author),
            noload(Post.author, User.posts),
            noload(Post.author, User.comments),
            joinedload(Post.comments, Comment.author),
            noload(Post.comments, Comment.author, User.posts),
            noload(Post.comments, Comment.author, User.comments),
            undefer(Post.title),
        )
        execution_options = {}
        if n_posts_in_subreddit >= MIN_POSTS_FOR_STREAM:
            execution_options["yield_per"] = POST_LIST_YIELD
        post_stmt = (
            select(Post)
            .where(
                Post.subreddit_name == subreddit.name,
            )
            .order_by(
                {
                    "top": desc(Post.score),
                    "new": desc(Post.created_utc),
                }[task.subtask]
            )
            .options(
                undefer(Post.title),
                undefer(Post.url),
                *options,
            )
            .execution_options(
                **execution_options,
            )
        )
        posts = self.session.scalars(post_stmt)
        self.log("Rendering subreddit...")
        result = self.renderer.render_subreddit(
            subreddit=subreddit,
            posts=posts,
            num_posts=n_posts_in_subreddit,
            sort=task.subtask,
        )
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_subreddit_wiki_task(self, task):
        """
        Process a received subreddit wiki task.

        Note: this is called from L{Worker.process_subreddit_task}.

        @param task: task to process
        @type task: L{SubredditRenderTask}
        """
        # load subreddit
        self.log("Loading subreddit...")
        subreddit_stmt = (
            select(Subreddit)
            .where(
                Subreddit.name == task.subreddit_name,
            )
            .options(
                raiseload(Subreddit.posts),
                raiseload(Subreddit.comments),
                selectinload(Subreddit.wikipages),
                undefer(Subreddit.wikipages, WikiPage.path),
                undefer(Subreddit.wikipages, WikiPage.revision_reason),
                undefer(Subreddit.wikipages, WikiPage.content),
            )
        )
        subreddit = self.session.scalars(subreddit_stmt).first()
        self.session.expunge(subreddit)  # prevent subreddit from being modified and storing objects
        self.log("Subreddit loaded.")
        if subreddit is None:
            self.log("-> Subreddit not found!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return

        if not subreddit.wikipages:
            self.log("-> Subreddit has no wiki!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return

        self.log("Rendering subreddit wiki...")
        result = self.renderer.render_subreddit_wiki(
            subreddit=subreddit,
        )
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_subreddit_rules_task(self, task):
        """
        Process a received subreddit rules task.

        Note: this is called from L{Worker.process_subreddit_task}.

        @param task: task to process
        @type task: L{SubredditRenderTask}
        """
        # load subreddit
        self.log("Loading subreddit...")
        subreddit_stmt = (
            select(Subreddit)
            .where(
                Subreddit.name == task.subreddit_name,
            )
            .options(
                raiseload(Subreddit.posts),
                raiseload(Subreddit.comments),
                selectinload(Subreddit.wikipages),
                selectinload(Subreddit.rules),
                undefer(Subreddit.rules, SubredditRule.short_name),
                undefer(Subreddit.rules, SubredditRule.description),
                undefer(Subreddit.rules, SubredditRule.violation_reason),
            )
        )
        subreddit = self.session.scalars(subreddit_stmt).first()
        self.session.expunge(subreddit)  # prevent subreddit from being modified and storing objects
        self.log("Subreddit loaded.")
        if subreddit is None:
            self.log("-> Subreddit not found!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return

        self.log("Rendering subreddit rules...")
        result = self.renderer.render_subreddit_rules(
            subreddit=subreddit,
        )
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_subreddit_stats_task(self, task):
        """
        Process a received subreddit statistics task.

        Note: this is called from L{Worker.process_subreddit_task}.

        @param task: task to process
        @type task: L{SubredditRenderTask}
        """
        # load subreddit
        self.log("Loading subreddit...")
        subreddit_stmt = (
            select(Subreddit)
            .where(
                Subreddit.name == task.subreddit_name,
            )
            .options(
                raiseload(Subreddit.posts),
                raiseload(Subreddit.comments),
                joinedload(Subreddit.wikipages),
            )
        )
        subreddit = self.session.scalars(subreddit_stmt).first()
        self.session.expunge(subreddit)  # prevent subreddit from being modified and storing objects
        self.log("Subreddit loaded.")
        if subreddit is None:
            self.log("-> Subreddit not found!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return

        # collect stats
        self.log("Gathering statistics")
        stats = query_post_stats(
            session=self.session,
            post_filter=(Post.subreddit_name == task.subreddit_name),
            comment_filter=(Comment.subreddit_name == task.subreddit_name),
        )

        self.log("Rendering subreddit statistics...")
        result = self.renderer.render_subreddit_stats(
            subreddit=subreddit,
            stats=stats,
        )
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_user_task(self, task):
        """
        Process a received user task.

        @param task: task to process
        @type task: L{SubredditRenderTask}
        """
        if task.part == "posts":
            self.process_user_posts_task(task)
        elif task.part == "comments":
            self.process_user_comments_task(task)
        elif task.part == "stats":
            self.process_user_stats_task(task)
        else:
            raise ValueError("Unknown user part: {}".format(task.part))

    def process_user_posts_task(self, task):
        """
        Process the posts of a user.

        @param task: task to process
        @type task: L{UserRenderTask}
        """
        # load user
        self.log("Loading user...")
        user_stmt = (
            select(User)
            .where(
                User.name == task.username,
            )
            .options(
                raiseload(User.posts),
                raiseload(User.comments),
            )
        )
        user = self.session.scalars(user_stmt).first()
        if user is None:
            self.log("-> User not found!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return
        self.session.expunge(user)  # prevent user from being modified and storing objects
        user_name = user.name
        self.log("User loaded. Username is: {}".format(user_name))
        # count posts in subreddit
        self.log("Counting posts made by user...")
        count_stmt = (
            select(func.count(Post.uid))
            .where(
                Post.author_name == user_name,
            )
        )
        n_posts_by_user = self.session.execute(count_stmt).scalar_one()
        self.log("Found {} posts.".format(n_posts_by_user))

        # load posts
        self.log("Starting to load posts...")
        # always use eager loading, lazy is horrible for performance here
        options = (
            selectinload(Post.comments),
            selectinload(Post.author),
            # noload(Post.author, User.posts),
            # noload(Post.author, User.comments),
            noload(Post.comments, Comment.author),
            noload(Post.comments, Comment.author, User.posts),
            noload(Post.comments, Comment.author, User.comments),
        )
        execution_options = {}
        if n_posts_by_user >= MIN_POSTS_FOR_STREAM:
            execution_options["yield_per"] = POST_LIST_YIELD
        post_stmt = (
            select(Post)
            .where(
                Post.author_name == user_name,
            )
            .order_by(
                {
                    "top": desc(Post.score),
                    "new": desc(Post.created_utc),
                }[task.sort]
            )
            .options(
                undefer(Post.title),
                undefer(Post.url),
                *options,
            )
            .execution_options(
                **execution_options,
            )
        )
        posts = self.session.scalars(post_stmt)
        self.log("Rendering user posts...")
        result = self.renderer.render_user_posts(
            user=user,
            posts=posts,
            num_posts=n_posts_by_user,
            sort=task.sort,
        )
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_user_comments_task(self, task):
        """
        Process the comments of a user.

        @param task: task to process
        @type task: L{UserRenderTask}
        """
        # load user
        self.log("Loading user...")
        user_stmt = (
            select(User)
            .where(
                User.name == task.username,
            )
            .options(
                raiseload(User.posts),
                raiseload(User.comments),
            )
        )
        user = self.session.scalars(user_stmt).first()
        if user is None:
            self.log("-> User not found!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return
        self.session.expunge(user)  # prevent user from being modified and storing objects
        user_name = user.name
        self.log("User loaded. Username is: {}".format(user_name))
        # count posts in subreddit
        self.log("Counting comments made by user...")
        count_stmt = (
            select(func.count(Comment.uid))
            .where(
                Comment.author_name == user_name,
            )
        )
        n_comments_by_user = self.session.execute(count_stmt).scalar_one()
        self.log("Found {} comments.".format(n_comments_by_user))

        # load posts
        self.log("Starting to load comments...")
        # always use eager loading, lazy is horrible for performance here
        options = (
            selectinload(Comment.author),
            joinedload(Comment.post),
            joinedload(Comment.post, Post.author),
            selectinload(Comment.post, Post.comments),
            selectinload(Comment.post, Post.subreddit),
            joinedload(Comment.subreddit),
            noload(Comment.subreddit, Subreddit.posts),
            noload(Comment.subreddit, Subreddit.comments),
        )
        execution_options = {}
        if n_comments_by_user >= MIN_POSTS_FOR_STREAM:
            execution_options["yield_per"] = POST_LIST_YIELD
        comment_stmt = (
            select(Comment)
            .where(
                Comment.author_name == user_name,
            )
            .order_by(
                {
                    "top": desc(Comment.score),
                    "new": desc(Comment.created_utc),
                }[task.sort]
            )
            .options(
                undefer(Comment.body),
                *options,
            )
            .execution_options(
                **execution_options,
            )
        )
        comments = self.session.scalars(comment_stmt)
        self.log("Rendering user comments...")
        result = self.renderer.render_user_comments(
            user=user,
            comments=comments,
            num_comments=n_comments_by_user,
            sort=task.sort,
        )
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_user_stats_task(self, task):
        """
        Process a received user statistics task.

        Note: this is called from L{Worker.process_user_task}.

        @param task: task to process
        @type task: L{UserRenderTask}
        """
        # load user
        self.log("Loading user...")
        user_stmt = (
            select(User)
            .where(
                User.name == task.username,
            )
            .options(
                raiseload(User.posts),
                raiseload(User.comments),
            )
        )
        user = self.session.scalars(user_stmt).first()
        if user is None:
            self.log("-> User not found!")
            self.log("Submitting empty result...")
            result = RenderResult()
            self.handle_result(result)
            self.log("Done.")
            return
        self.session.expunge(user)  # prevent user from being modified and storing objects
        user_name = user.name
        self.log("User loaded. Username is: {}".format(user_name))

        # collect stats
        self.log("Gathering statistics")
        stats = query_post_stats(
            session=self.session,
            post_filter=(Post.author_name == user_name),
            comment_filter=(Comment.author_name == user_name),
        )

        self.log("Rendering user statistics...")
        result = self.renderer.render_user_stats(
            user=user,
            stats=stats,
        )
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_etc_task(self, task):
        """
        Process a received etc task.

        @param task: task to process
        @type task: L{EtcRenderTask}
        """
        if task.subtask == "index":
            # render the indexpage
            result = self.process_index_page_task(task)
        elif task.subtask == "subreddits":
            # render the list of subreddits
            result = self.process_subreddit_list_task(task)
        elif task.subtask == "stats":
            # render the global statistics
            self.log("Querying stats...")
            stats = query_post_stats(self.session, True, True)
            self.log("Rendering statistics...")
            result = self.renderer.render_global_stats(stats)
        elif task.subtask == "scripts":
            # the js scripts
            self.log("Including js scripts...")
            result = self.renderer.render_scripts()
        elif task.subtask == "info":
            # the info pages
            self.log("Rendering info pages...")
            result = self.renderer.render_info_pages()
        else:
            raise ValueError("Unknown etc subtask: '{}'!".format(task.subtask))
        self.log("Submitting result...")
        self.handle_result(result)
        self.log("Done.")

    def process_index_page_task(self, task):
        """
        Process a received task for the index page.

        Note: this is called from L{Worker.process_etc_task}.

        @param task: task to process
        @type task: L{EtcRenderTask}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        self.log("Rendering index page...")
        subreddit_infos = []
        post_stmt = (
            select(Post.subreddit_name, func.count(Post.uid).label("count"))
            .group_by(Post.subreddit_name)
            .order_by(desc("count"))
            .limit(SUBREDDITS_ON_INDEX_PAGE)
        )
        for subname, count in self.session.execute(post_stmt):
            subreddit_infos.append(SubredditInfo(subname, posts=count))
        result = self.renderer.render_index(subreddit_infos=subreddit_infos)
        return result

    def process_subreddit_list_task(self, task):
        """
        Process a received task for the subreddit list page.

        Note: this is called from L{Worker.process_etc_task}.

        @param task: task to process
        @type task: L{EtcRenderTask}
        @return: the rendered pages and redirects
        @rtype: L{RenderResult}
        """
        self.log("Rendering subreddit list...")
        subreddit_infos = []
        post_stmt = (
            select(Post.subreddit_name, func.count(Post.uid).label("count"))
            .group_by(Post.subreddit_name)
            .order_by(Post.subreddit_name)
        )
        for subname, count in self.session.execute(post_stmt):
            subreddit_infos.append(SubredditInfo(subname, posts=count))
        result = self.renderer.render_subreddit_list(subreddit_infos=subreddit_infos)
        return result



"""
This module manages the build of the ZIM files.

It handles the ZIM creator, add basic content, instantiates the workers,
issues them their tasks and adds the result to the creator.

@var MAX_OUTSTANDING_TASKS: max size of the task queue
@type MAX_OUTSTANDING_TASKS: L{int}
@var MAX_RESULT_BACKLOG: max size of the task result queue
@type MAX_RESULT_BACKLOG: L{int}
@var POSTS_PER_TASK: number of post IDs to send per worker tasks
@type POSTS_PER_TASK: L{int}
"""
import multiprocessing
import threading
import queue
import datetime
import time
import os
import contextlib
import math
import pdb
import signal
import pathlib

from scss.compiler import Compiler as ScssCompiler
from scss.namespace import Namespace as ScssNamespace
from scss.types import String as ScssString
from sqlalchemy import select, func
from sqlalchemy.orm import Session
from libzim.writer import Creator, Item, StringProvider, FileProvider, Hint
import tqdm

try:
    import psutil
except ImportError:
    psutil = None
try:
    import setproctitle
except ImportError:
    setproctitle = None

from ..util import get_package_dir, get_resource_file_path, set_or_increment
from ..util import format_timedelta, format_size, format_number, trim_title
from ..downloader import hash_url
from ..imgutils import mimetype_is_image, mimetype_is_video
from ..db.models import Post, Subreddit, User, MediaFile, ARCTICZIM_USERNAME
from .renderer import HtmlPage, Redirect, JsonObject, Script, FileReferences, RenderOptions
from .worker import Worker, WorkerOptions
from .worker import StopTask, PostRenderTask, EtcRenderTask, SubredditRenderTask, UserRenderTask
from .worker import MARKER_TASK_COMPLETED, MARKER_WORKER_STOPPED
from .buckets import BucketMaker


MAX_OUTSTANDING_TASKS = 1024 * 8
MAX_RESULT_BACKLOG = 1024
POSTS_PER_TASK = 64


# =============== HELPER FUNCTIONS ================


def get_n_cores():
    """
    Return the number of cores to use.
    If multiprocessing is available, this is the number of cores available.
    Otherwise, this will be 1.

    @return: the number of cores to use.
    @rtype: L{int}
    """
    if multiprocessing is not None:
        return multiprocessing.cpu_count()
    else:
        return 1


def config_process(name, nice=0, ionice=0):
    """
    Configure the current OS process.

    This function expects the linux values and will try to guess the
    approximate windows values.

    @param name: name for the current process
    @type name: L{str}
    @param nice: new nice value for current process (-21->19 (lowest))
    @type nice: L{int}
    @param ionice: new io nice value for the process (0->17 (lowest))
    @type ionice: L{int}
    """
    # name
    if setproctitle is not None:
        setproctitle.setproctitle(name)
    # nice and ionice
    if psutil is not None:
        p = psutil.Process()
        if psutil.LINUX:
            p.nice(nice)
            p.ionice(psutil.IOPRIO_CLASS_BE, ionice)
        else:
            if nice > 0:
                nv = psutil.ABOVE_NORMAL_PRIORITY_CLASS
            elif nice < 0:
                nv = psutil.BELOW_NORMAL_PRIORITY_CLASS
            else:
                nv = psutil.NORMAL_PRIORITY_CLASS
            p.nice(nv)
            if ionice < 4:
                iv = psutil.IOPRIO_HIGH
            elif ionice > 4:
                iv = psutil.IOPRIO_LOW
            else:
                iv = psutil.IOPRIO_NORMAL
            p.ionice(iv)


def config_thread(name):
    """
    Configure the current OS thread.

    @param name: new name of the thread
    @type name: L{str}
    """
    if setproctitle is not None:
        setproctitle.setthreadtitle(name)


# ================ DEBUG HELPER ============


def on_pdb_interrupt(sig, frame):
    """
    Called on an SIGUSR1 interrupt to start pdb debugging.
    """
    pdb.Pdb().set_trace(frame)


try:
    signal.signal(signal.SIGUSR1, on_pdb_interrupt)
except Exception:
    pass


# =============== ITEM DEFINITIONS ================


class HtmlPageItem(Item):
    """
    A L{libzim.writer.Item} for HTML pages.
    """
    def __init__(self, path, title, content, is_front=True):
        """
        The default constructor.

        @param path: path of the page in the ZIM file
        @type path: L{str}
        @param title: title of the page
        @type title: L{str}
        @param content: the content of the page
        @type content: L{str}
        @param is_front: if this is nonzero, set this as a front article
        @type is_front: L{bool}
        """
        super().__init__()
        self._path = path
        self._title = title
        self._content = content
        self._is_front = is_front

    def get_path(self):
        return self._path

    def get_title(self):
        return trim_title(self._title)

    def get_mimetype(self):
        return "text/html"

    def get_contentprovider(self):
        return StringProvider(self._content)

    def get_hints(self):
        return {
            Hint.FRONT_ARTICLE: self._is_front,
            Hint.COMPRESS: True,
        }


class JsonItem(Item):
    """
    A L{libzim.writer.Item} for json.
    """
    def __init__(self, path, title, content):
        """
        The default constructor.

        @param path: path of to store item in ZIM file
        @type path: L{str}
        @param title: title of the json file
        @type title: L{str}
        @param content: the content of the json file
        @type content: L{str}
        """
        super().__init__()
        self._path = path
        self._title = title
        self._content = content

    def get_path(self):
        return self._path

    def get_title(self):
        return self._title

    def get_mimetype(self):
        return "application/json"

    def get_contentprovider(self):
        return StringProvider(self._content)

    def get_hints(self):
        return {
            Hint.FRONT_ARTICLE: False,
            Hint.COMPRESS: True,
        }


class ScriptItem(Item):
    """
    A L{libzim.writer.Item} for a js script.
    """
    def __init__(self, path, title, content):
        """
        The default constructor.

        @param path: path of to store item in ZIM file
        @type path: L{str}
        @param title: title of the js file
        @type title: L{str}
        @param content: the content of the js file
        @type content: L{str}
        """
        super().__init__()
        self._path = path
        self._title = title
        self._content = content

    def get_path(self):
        return self._path

    def get_title(self):
        return self._title

    def get_mimetype(self):
        return "text/javascript"

    def get_contentprovider(self):
        return StringProvider(self._content)

    def get_hints(self):
        return {
            Hint.FRONT_ARTICLE: False,
            Hint.COMPRESS: True,
        }


class StylesheetItem(Item):
    """
    A L{libzim.writer.Item} for the CSS stylesheet.
    """
    def __init__(self, theme="light"):
        """
        The default constructor.

        @param theme: theme to render
        @type theme: L{str}
        """
        super().__init__()
        self._theme = theme

    def get_path(self):
        return "style_{}.css".format(self._theme)

    def get_title(self):
        return "CSS Stylesheet"

    def get_mimetype(self):
        return "text/css"

    def get_contentprovider(self):
        path = get_resource_file_path("style.scss")
        with open(path, "r") as fin:
            scss = fin.read()
        namespace = ScssNamespace()
        namespace.set_variable("$theme", ScssString(self._theme))
        compiler = ScssCompiler(
            root=pathlib.Path(get_package_dir()),
            search_path=["resources"],
            live_errors=False,  # raise exception when compilation fails
            namespace=namespace,
        )
        css = compiler.compile_string(scss)
        return StringProvider(css)

    def get_hints(self):
        return {
            Hint.FRONT_ARTICLE: False,
            Hint.COMPRESS: True,
        }


class FaviconItem(Item):
    """
    A L{libzim.writer.Item} for the favicon.
    """
    def __init__(self):
        """
        The default constructor.
        """
        super().__init__()

    def get_path(self):
        return "favicon.png"

    def get_title(self):
        return "Favicon (PNG)"

    def get_mimetype(self):
        return "image/png"

    def get_contentprovider(self):
        return FileProvider(get_resource_file_path("icons", "arcticzim_highres.png"))

    def get_hints(self):
        return {
            Hint.FRONT_ARTICLE: False,
            Hint.COMPRESS: True,
        }


class IconItem(Item):
    """
    A L{libzim.writer.Item} for various icons.
    """
    def __init__(self, icon_name):
        """
        The default constructor.

        @param icon_name: name of the icon to include
        @type icon_name: L{str}
        """
        super().__init__()
        self.icon_name = icon_name

    def get_path(self):
        return "icons/icon_{}.png".format(self.icon_name)

    def get_title(self):
        return "Icon for {}".format(self.icon_name)

    def get_mimetype(self):
        return "image/png"

    def get_contentprovider(self):
        return FileProvider(get_resource_file_path("icons", "icon_{}.png".format(self.icon_name)))

    def get_hints(self):
        return {
            Hint.FRONT_ARTICLE: False,
            Hint.COMPRESS: True,
        }


class MediaItem(Item):
    """
    A L{libzim.writer.Item} for media files.
    """
    def __init__(self, mediadir, mediafile):
        """
        The default constructor.

        @param mediadir: path to the media directory
        @type mediadir: L{str}
        @param mediafile: the media file this is for
        @type mediafile: L{arcticzim.db.models.MediaFile}
        """
        self._mediadir = mediadir
        self._uid = mediafile.uid
        self._mimetype = mediafile.mimetype
        self._url = mediafile.url

    def get_path(self):
        return "media/{}".format(self._uid)

    def get_title(self):
        return self._url

    def get_mimetype(self):
        return self._mimetype

    def get_contentprovider(self):
        return FileProvider(os.path.join(self._mediadir, hash_url(self._url)))

    def get_hints(self):
        return {
            Hint.FRONT_ARTICLE: False,
            Hint.COMPRESS: True,
        }


# =============== BUILD LOGIC =================


class BuildOptions(object):
    """
    A class containing the build options for the ZIM.

    @ivar name: human-readable identifier of the resource
    @type name: L{str}
    @ivar title: title of ZIM file
    @type title: L{str}
    @ivar creator: creator of the ZIM file content
    @type creator: L{str}
    @ivar publisher: publisher of the ZIM file
    @type publisher: L{str}
    @ivar description: description of the ZIM file
    @type description: L{str}
    @ivar language: language to use (e.g. "eng")
    @type language: L{str}
    @ivar indexing: whether indexing should be enabled or not
    @type indexing: L{bool}

    @ivar with_stats: if nonzero, include statistics
    @type with_stats: L{bool}
    @ivar with_users: if nonzero, include user pages
    @type with_users: L{bool}
    @ivar with_media: if nonzero, include media files
    @type with_media: L{str}
    @ivar with_videos: if nonzero, include video files
    @type with_videos: L{str}

    @ivar use_threads: if nonzero, use threads instead of processes
    @type use_threads: L{bool}
    @ivar num_workers: number of (non-zim) workers to use
    @type num_workers: L{int}
    @ivar log_directory: if not None, enable logging and write logs into this directory
    @type log_directory: L{str} or L{None}

    @ivar eager: if nonzero, eager load objects from database
    @type eager: L{bool}
    @ivar memprofile_directory: if not None, enable memory profiling and write files into this directory
    @type memprpofile_directory: L{str} or L{None}

    @ivar skip_posts: debug option to not render posts
    @type skip_posts: L{bool}
    """
    def __init__(
        self,

        # ZIM options
        name="arcticzim_eng",
        title="ArcticZim",
        creator="Reddit and ArcticShift",
        publisher="ArcticZim",
        description="ZIM file containing a part of reddit",
        language="eng",
        indexing=True,

        # content options
        with_stats=True,
        with_users=True,
        with_media=True,
        with_videos=False,

        # genral build_options
        log_directory=None,

        # worker management options
        use_threads=False,
        num_workers=None,

        # worker options
        eager=True,
        memprofile_directory=None,

        # debug options
        skip_posts=False,
    ):
        """
        The default constructor.

        @param name: human-readable identifier of the resource
        @type name: L{str}
        @param title: title of ZIM file
        @type title: L{str}
        @param creator: creator of the ZIM file content
        @type creator: L{str}
        @param publisher: publisher of the ZIM file
        @type publisher: L{str}
        @param description: description of the ZIM file
        @type description: L{str}
        @param language: language to use (e.g. "eng")
        @type language: L{str}
        @param indexing: whether indexing should be enabled or not
        @type indexing: L{bool}

        @param use_threads: if nonzero, use threads instead of processes
        @type use_threads: L{bool}
        @param num_workers: number of (non-zim) workers to use (None -> auto)
        @type num_workers: L{int} or L{None}

        @param log_directory: if specified, enable logging and write logs into this directory
        @type log_directory: L{str} or L{None}

        @param eager: if nonzero, eager load objects from database
        @type eager: L{bool}
        @param memprofile_directory: if specified, enable memory profiling and write files into this directory
        @type memprofile_directory: L{str} or L{None}

        @param skip_posts: debug option to not render posts
        @type skip_posts: L{bool}
        @param with_stats: if nonzero, include statistics
        @type with_stats: L{bool}
        @param with_users: if nonzero, include user pages
        @type with_users: L{bool}
        @param with_media: if nonzero, include media files
        @type with_media: L{str}
        @param with_videos: if nonzero, include video files
        @type with_videos: L{str}
        """
        self.name = name
        self.title = title
        self.creator = creator
        self.publisher = publisher
        self.description = description
        self.language = language
        self.indexing = indexing

        self.with_stats = with_stats
        self.with_users = with_users
        self.with_media = with_media
        self.with_videos = with_videos

        self.use_threads = bool(use_threads)
        if num_workers is None:
            self.num_workers = get_n_cores()
        else:
            self.num_workers = int(num_workers)

        self.log_directory = log_directory

        self.eager = eager
        self.memprofile_directory = memprofile_directory

        self.skip_posts = skip_posts

    @staticmethod
    def add_argparse_options(parser):
        """
        Add all CLI options to the specified argparse parser.

        @param parser: argument parser to which to add the arguments
        @type parser: L{argparse.ArgumentParser}
        """
        parser.add_argument(
            "--name",
            action="store",
            dest="name",
            default="arcticzim_eng",
            help="a human readable identifier for the ZIM",
        )
        parser.add_argument(
            "--title",
            action="store",
            dest="title",
            default="ArcticZim",
            help="the title of the ZIM file",
        )
        parser.add_argument(
            "--creator",
            action="store",
            dest="creator",
            default="Reddit and Arctic Shift",
            help="creator(s) of the ZIM file content",
        )
        parser.add_argument(
            "--publisher",
            action="store",
            dest="publisher",
            default="ArcticZim",
            help="creator of the ZIM file itself",
        )
        parser.add_argument(
            "--description",
            action="store",
            dest="description",
            default="A ZIM file containing a part of reddit",
            help="a short description of the content",
        )
        parser.add_argument(
            "--language",
            action="store",
            dest="language",
            default="eng",
            help="ISO639-3 language identifier describing content language",
        )
        parser.add_argument(
            "--no-indexing",
            action="store_false",
            dest="indexing",
            help="disable indexing of ZIM",
        )

        parser.add_argument(
            "--threaded",
            action="store_true",
            help="use threads instead of processes for workers"
        )
        parser.add_argument(
            "--workers",
            action="store",
            type=int,
            default=None,
            help="use this many non-zim workers",
        )
        parser.add_argument(
            "--log-directory",
            action="store",
            default=None,
            help="enable logging and write logs into this directory",
        )
        parser.add_argument(
            "--no-stats",
            action="store_false",
            dest="with_stats",
            help="do not include statistics",
        )
        parser.add_argument(
            "--no-media",
            action="store_false",
            dest="with_media",
            help="do not include media",
        )
        parser.add_argument(
            "--with-videos",
            action="store_true",
            dest="with_videos",
            help="include videos",
        )
        parser.add_argument(
            "--no-users",
            action="store_false",
            dest="with_users",
            help="do not include user pages",
        )
        parser.add_argument(
            "--lazy",
            action="store_false",
            dest="eager",
            help="Do not eager load related objects, ...",
        )
        parser.add_argument(
            "--memprofile-directory",
            action="store",
            default=None,
            help="enable memory profile and write into this directory",
        )
        parser.add_argument(
            "--debug-skip-posts",
            action="store_true",
            dest="skip_posts",
            help="do not include posts (debug option)",
        )

    @classmethod
    def from_ns(cls, ns):
        """
        Instantiate build options from a namespace returned by a parser.
        This method assumes that the parser was preprared using L{BuildOptions.add_argparse_options}.

        @param ns: namespace containg the arguments
        @type ns: L{argparse.Namespace}
        """
        bo = cls(
            name=ns.name,
            title=ns.title,
            creator=ns.creator,
            publisher=ns.publisher,
            description=ns.description,
            language=ns.language,
            indexing=ns.indexing,

            use_threads=ns.threaded,
            num_workers=ns.workers,
            log_directory=ns.log_directory,
            eager=ns.eager,
            memprofile_directory=ns.memprofile_directory,

            with_stats=ns.with_stats,
            with_users=ns.with_users,
            with_media=ns.with_media,
            with_videos=ns.with_videos,
            skip_posts=ns.skip_posts,
        )
        return bo

    def get_metadata_dict(self):
        """
        Return a dictionary encoding the ZIM metadata described by this file.

        Additional metadata will likely be added.

        @return: a dictionary containing the metadata of this ZIM file.
        @rtype: L{bool}
        """
        tags = [
            "_sw:no",
            "_ftindex:" + ("yes" if self.indexing else "no"),
            "_pictures:" + ("yes" if self.with_media else "no"),
            "_videos:" + ("yes" if self.with_videos else "no"),
            "_category:reddit",
        ]
        metadata = {
            "Name": self.name,
            "Title": self.title,
            "Creator": self.creator,
            "Date": datetime.date.today().isoformat(),
            "Publisher": self.publisher,
            "Description": self.description,
            "Language": self.language,
            "Tags": ";".join(tags),
            "Scraper": "arcticzim",
        }
        return metadata

    def get_worker_options(self):
        """
        Return the worker options the worker should use.

        @return: options for the worker.
        @rtype: L{arcticzim.zimbuild.worker.WorkerOptions}
        """
        options = WorkerOptions(
            eager=self.eager,
            memprofile_directory=self.memprofile_directory,
            log_directory=self.log_directory,
            with_stats=self.with_stats,
            with_media=self.with_media,
            with_videos=self.with_videos,
        )
        return options

    def get_render_options(self):
        """
        Return the render options the renderer should use.

        @return: options for the renderer
        @rtype: L{arcticzim.zimbuild.renderer.RenderOptions}
        """
        options = RenderOptions(
            with_stats=self.with_stats,
            with_users=self.with_users,
            with_videos=self.with_videos,
        )
        return options


class ZimBuilder(object):
    """
    The ZimBuilder manages the ZIM build process.

    @ivar inqueue: the queue where tasks will be put
    @type inqueue: L{multiprocessing.Queue} or L{queue.Queue}
    @ivar outqueue: the queue containing the task results
    @type outqueue: L{multiprocessing.Queue} or L{queue.Queue}
    @ivar connection_config: configuration for database connection
    @type connection_config: L{arcticzim.db.connection.ConnectionConfig}
    @ivar num_files_added: a dict mapping a filetype to the number of files of that type
    @type num_files_added: L{dict} of L{str} -> L{int}
    @ivar next_worker_id: ID to give next worker
    @type next_worker_id: L{int}
    @ivar mediadir: location where media files are stored
    @type mediadir: L{str}
    """
    def __init__(self, connection_config, mediadir):
        """
        The default constructor.

        @param connection_config: configuration for database connection
        @type connection_config: L{arcticzim.db.connection.ConnectionConfig}
        @param mediadir: location where media files are stored
        @type mediadir: L{str}
        """
        self.connection_config = connection_config

        self.inqueue = None
        self.outqueue = None
        self.num_files_added = {}
        self.media_file_references = set()
        self.next_worker_id = 0
        self.mediadir = mediadir

    def _init_queues(self, options):
        """
        Initialize the queues.

        @param options: build options
        @type options: L{BuildOptions}
        """
        if options.use_threads:
            self.inqueue = queue.Queue(maxsize=MAX_OUTSTANDING_TASKS)
            self.outqueue = queue.Queue(maxsize=MAX_RESULT_BACKLOG)
        else:
            self.inqueue = multiprocessing.Queue(maxsize=MAX_OUTSTANDING_TASKS)
            self.outqueue = multiprocessing.Queue(maxsize=MAX_RESULT_BACKLOG)

    def cleanup(self):
        """
        Perform clean up tasks.
        """
        pass
        # self.session.close()
        # self.engine.dispose()

    def log(self, msg, end="\n"):
        """
        Log a message.

        @param msg: message to log
        @type msg: L{str}
        @param end: suffix for the message, defaults to newline
        @type end: L{str}
        """
        # note: this is primarily a simple replacement for historical functionality.
        print(msg, end=end)

    def build(self, outpath, options):
        """
        Build a ZIM.

        @param outpath: path to write ZIM to
        @type outpath: L{str}
        @param options: build options for the ZIM
        @type options: L{BuildOptions}
        """
        # prepare build
        self.log("Preparing build...")

        start = time.time()

        # find and report options for the build
        self.log(" -> Generating ZIM creation config...")
        compression = "zstd"
        # clustersize = 8 * 1024 * 1024  # 8 MiB
        clustersize = 2 * 1024 * 1024  # 2 MiB
        verbose = True
        n_creator_workers = get_n_cores()
        n_render_workers = options.num_workers
        use_threads = options.use_threads
        self.log("        -> Path:             {}".format(outpath))
        self.log("        -> Verbose:          {}".format(verbose))
        self.log("        -> Compression:      {}".format(compression))
        self.log("        -> Cluster size:     {}".format(format_size(clustersize)))
        self.log("        -> Creator Workers:  {}".format(n_creator_workers))
        self.log("        -> Render Workers:   {}".format(n_render_workers))
        self.log("            -> using: {}".format("threads" if use_threads else "processes"))
        if not use_threads:
            self.log("            -> started using: {}".format(multiprocessing.get_start_method()))
        self.log("            -> eagerloading: {}".format("enabled" if options.eager else "disabled"))
        self.log("        -> Done.")

        # connect to database
        # initially, we did this in __init__ and set the engine+session
        # as attributes, but as the engine is not pickable, this prevents
        # the use of "forkserver" for multiprocessing
        self.log(" -> Establishing database connection... ", end="")
        engine = self.connection_config.connect()
        session = Session(engine)
        self.log("Done.")

        # initailize queues
        self.log(" -> Initiating queues...", end="")
        self._init_queues(options)
        self.log("Done.")

        # open the ZIM creator
        self.log("Opening ZIM creator, writing to path '{}'... ".format(outpath), end="")
        with Creator(outpath) \
                .config_indexing(options.indexing, options.language) \
                .config_clustersize(clustersize) \
                .config_verbose(verbose) \
                .config_nbworkers(n_creator_workers) as creator:
            self.log("Done.")

            # configurations
            self.log("Configuring ZIM... ", end="")
            creator.set_mainpath("index.html")
            self.log("Done.")

            # add metadata
            self.log("Adding metadata... ", end="")
            metadata = options.get_metadata_dict()
            for key, value in metadata.items():
                creator.add_metadata(key, value)
                set_or_increment(self.num_files_added, "metadata")
            self.log("Done.")

            # add illustrations
            self.log("Adding main illustration... ", end="")
            imagepath = get_resource_file_path("icons", "arcticzim.png")
            with open(imagepath, "rb") as fin:
                creator.add_illustration(48, fin.read())
            set_or_increment(self.num_files_added, "image")
            creator.add_item(FaviconItem())
            set_or_increment(self.num_files_added, "image")
            for icon_name in ("text", "poll", "img", "link", "video"):
                creator.add_item(IconItem(icon_name))
                set_or_increment(self.num_files_added, "image")
            self.log("Done.")

            # add general items
            self.log("Adding stylesheet... ", end="")
            creator.add_item(StylesheetItem(theme="light"))
            set_or_increment(self.num_files_added, "css")
            creator.add_item(StylesheetItem(theme="dark"))
            set_or_increment(self.num_files_added, "css")
            self.log("Done.")

            # add content
            self._add_content(creator, session, options=options)

            # finish up
            self.log("Finalizing ZIM...")
        self.log("Done.")
        self.log("Cleaning up... ", end="")
        self.cleanup()
        self.log("Done.")

        # We're done, find and report some stats about the build
        final_size = os.stat(outpath).st_size
        end = time.time()
        time_elapsed = end - start
        self.log("Finished ZIM creation in {}.".format(format_timedelta(time_elapsed)))
        self.log("Final size: {}".format(format_size(final_size)))
        self.log("Added files: ")
        for filetype, amount in self.num_files_added.items():
            if filetype != "total":
                # print total later
                self.log("    {}: {} ({})".format(filetype, amount, format_number(amount)))
        total_amount = self.num_files_added["total"]
        self.log("    total: {} ({})".format(total_amount, format_number(total_amount)))

    def _add_content(self, creator, session, options):
        """
        Add the content of the ZIM file.

        @param creator: the ZIM creator
        @type creator: L{libzim.writer.Creator}
        @param session: sqlalchemy session for data querying
        @type session: L{sqlalchemy.orm.Session}
        @param options: build options for the ZIM
        @type options: L{BuildOptions}
        """
        self.log("Adding content...")
        # --- miscelaneous pages ---
        self.log(" -> Adding miscelaneous pages...")
        n_misc_pages = 3
        if options.with_stats:
            # account for the stat task
            n_misc_pages += 1
        with self._run_stage(
            creator=creator,
            options=options,
            task_name="Adding miscelaneous pages",
            n_tasks=n_misc_pages,
            task_unit="pages",
        ):
            self._send_etc_tasks(session, with_stats=options.with_stats)
        # --- subreddits ---
        self.log(" -> Adding subreddits...")
        self.log("     -> Finding subreddits... ", end="")
        n_subreddits = session.execute(
            select(func.count(Subreddit.name))
        ).scalar_one()
        self.log("found {} subreddits.".format(n_subreddits))
        n_tasks_per_subreddit = 4
        if options.with_stats:
            # one extra task for the stats
            n_tasks_per_subreddit += 1
        with self._run_stage(
            creator=creator,
            options=options,
            task_name="Adding subreddits",
            n_tasks=n_subreddits * n_tasks_per_subreddit,
            task_multiplier=(1 / n_tasks_per_subreddit),
            task_unit="subreddits",
        ):
            self._send_subreddit_tasks(session, with_stats=options.with_stats)
        # --- users ---
        if options.with_users:
            self.log(" -> Adding users...")
            self.log("     -> Finding users... ", end="")
            n_users = session.execute(
                select(func.count(User.name)).where(User.name != ARCTICZIM_USERNAME)
            ).scalar_one()
            self.log("found {} users.".format(n_users))
            n_tasks_per_user = 4
            if options.with_stats:
                # again, extra task for the stats
                n_tasks_per_user += 1
            with self._run_stage(
                creator=creator,
                options=options,
                task_name="Adding users",
                n_tasks=n_users * n_tasks_per_user,  # x tasks per user
                task_multiplier=(1 / n_tasks_per_user),
                task_unit="users",
            ):
                self._send_user_tasks(session, with_stats=options.with_stats)
        else:
            self.log(" -> Skipping users!")
        # --- posts ---
        if not options.skip_posts:
            self.log(" -> Adding posts...")
            self.log("     -> Finding posts... ", end="")
            n_posts = session.execute(
                select(func.count(Post.uid))
            ).scalar_one()
            self.log("found {} posts.".format(n_posts))
            n_post_tasks = math.ceil(n_posts / POSTS_PER_TASK)
            with self._run_stage(
                creator=creator,
                options=options,
                task_name="Adding posts",
                n_tasks=n_post_tasks,
                task_unit="posts",
                task_multiplier=POSTS_PER_TASK,
            ):
                self._send_post_tasks(session)
        else:
            self.log(" -> Skipping posts!")
        # --- media ---
        if options.with_media:
            self.log(" -> Adding media...")
            n = len(self.media_file_references)
            with tqdm.tqdm(desc="Adding files", total=n, unit="files") as bar:
                while self.media_file_references:
                    uid = self.media_file_references.pop()
                    stmt = select(MediaFile).where(MediaFile.uid == uid)
                    mf = session.execute(stmt).one()[0]
                    item = MediaItem(self.mediadir, mf)
                    creator.add_item(item)
                    if mimetype_is_image(mf.mimetype):
                        set_or_increment(self.num_files_added, "image", 1)
                    elif mimetype_is_video(mf.mimetype):
                        set_or_increment(self.num_files_added, "video", 1)
                    else:
                        set_or_increment(self.num_files_added, "other media", 1)
                    bar.update(1)
        else:
            self.log(" -> Skipping media!")

    @contextlib.contextmanager
    def _run_stage(self, options, **kwargs):
        """
        Add the content of the ZIM file.

        @param options: zim build options
        @type options: L{BuildOptions}
        @param kwargs: keyword arguments passed to L{ZimBuilder._creator_thread}
        @type kwargs: L{dict}
        """

        if "options" not in kwargs:
            kwargs["options"] = options
        n_workers = options.num_workers

        worker_class = (threading.Thread if options.use_threads else multiprocessing.Process)
        worker_options = options.get_worker_options()
        render_options = options.get_render_options()

        # start workers
        self.log("     -> Starting workers... ", end="")
        workers = []
        for i in range(n_workers):
            worker_id = self.next_worker_id
            self.next_worker_id += 1
            worker = worker_class(
                name="Content worker {}".format(worker_id),
                target=self._worker_process,
                kwargs={
                    "id": worker_id,
                    "connection_config": self.connection_config,
                    "worker_options": worker_options,
                    "render_options": render_options,
                },
            )
            worker.daemon = True
            worker.start()
            workers.append(worker)
        self.log("Done.")

        # start the background creator thread
        # for the duration of this method, only the thread is allowed
        # to work with the creator directly
        self.log("     -> Starting creator thread... ", end="")
        creator_thread = threading.Thread(  # <-- always use threads here
            name="Creator content adder thread",
            target=self._creator_thread,
            kwargs=kwargs,
        )
        creator_thread.daemon = True
        creator_thread.start()
        self.log("Done.")

        # now it's finally time to add the tasks
        yield

        # finish up
        self.log("     -> Waiting for workers... ", end="")
        # put stop tasks on queue
        for i in range(n_workers):
            self.inqueue.put(StopTask())
        # join with all workers
        for worker in workers:
            worker.join()
            if hasattr(worker, "close"):
                worker.close()
        self.log("Done.")

        self.log("     -> Joining with creator thread... ", end="")
        creator_thread.join()
        self.log("Done.")
        self.log("     -> Done.")

    def _creator_thread(
        self,
        creator,
        options,
        task_name,
        n_tasks,
        task_unit,
        task_multiplier=1,
    ):
        """
        This method will be executed as the creator thread.

        This function is responsible to get results from the outqueue
        and adding them to the ZIM file.

        @param creator: creator for the ZIM file
        @type creator: L{libzim.writer.Creator}
        @param options: zim build options
        @type options: L{BuildOptions}
        @param taskname
        @param task_name: name of the task that is currently being processed
        @type task_name: L{str}
        @param n_tasks: number of tasks issued
        @type n_tasks: L{int}
        @param task_unit: string describing the unit of the task (e.g. stories)
        @type task_unit: L{str}
        @param task_multiplier: multiply bar advancement per task by this factor
        @type task_multiplier: L{int}
        """
        if not options.use_threads:
            # setup priority first
            config_process(name="AZ creator", nice=2, ionice=7)
        config_thread(name="Creator thread")
        # main loop - get results from queue and add them to ZIM
        running = True
        n_finished = 0
        n_items_added = 0
        with tqdm.tqdm(
            desc=task_name,
            total=n_tasks * task_multiplier,
            unit=task_unit,
            # custom bar format to deal with floats
            # bar_format="{l_bar}{bar}| {n_fmt:.2f}/{total_fmt:.2f} [{elapsed:}<{remaining:.}, {rate_fmt}{postfix}]",
            postfix={"items": n_items_added},
        ) as bar:
            while running:
                render_result = self.outqueue.get(block=True)
                bar.refresh()
                if render_result == MARKER_WORKER_STOPPED:
                    # worker finished
                    n_finished += 1
                    if n_finished == options.num_workers:
                        # all workers shut down
                        running = False
                elif render_result == MARKER_TASK_COMPLETED:
                    # task was completed
                    bar.update(task_multiplier)
                else:
                    # add the rendered objects to the ZIM
                    for rendered_object in render_result.iter_objects():
                        if isinstance(rendered_object, HtmlPage):
                            # add a HTML page
                            item = HtmlPageItem(
                                path=rendered_object.path,
                                title=rendered_object.title,
                                content=rendered_object.content,
                                is_front=rendered_object.is_front,
                            )
                            creator.add_item(item)
                            set_or_increment(self.num_files_added, "html")
                        elif isinstance(rendered_object, JsonObject):
                            # add a json object
                            item = JsonItem(
                                path=rendered_object.path,
                                title=rendered_object.title,
                                content=rendered_object.content,
                            )
                            creator.add_item(item)
                            set_or_increment(self.num_files_added, "json")
                        elif isinstance(rendered_object, Redirect):
                            # create a redirect
                            creator.add_redirection(
                                rendered_object.source,
                                rendered_object.title,
                                rendered_object.target,
                                hints={
                                    Hint.FRONT_ARTICLE: rendered_object.is_front,
                                }
                            )
                            set_or_increment(self.num_files_added, "redirect")
                        elif isinstance(rendered_object, Script):
                            # add a script
                            item = ScriptItem(
                                path=rendered_object.path,
                                title=rendered_object.title,
                                content=rendered_object.content,
                            )
                            creator.add_item(item)
                            set_or_increment(self.num_files_added, "js")
                        elif isinstance(rendered_object, FileReferences):
                            # register the file references
                            if options.with_media:
                                self.media_file_references.update(rendered_object.uids)
                        else:
                            # unknown result object
                            raise RuntimeError("Unknown render result: {}".format(type(rendered_object)))
                        set_or_increment(self.num_files_added, "total")
                        n_items_added += 1
                        bar.set_postfix({"items": n_items_added})

    def _worker_process(self, id, connection_config, worker_options, render_options):
        """
        This method will be executed as a worker process.

        The workers take tasks from the inqueue, process them and add
        the result to the outqueue.

        @param id: id of the worker
        @type id: L{int}
        @param connection_config: database connection configuration
        @type connection_config: L{arcticzim.db.connection.ConnectionConfig}
        @param worker_options: options for the worker
        @type worker_options: L{arcticzim.zimbuild.worker.WorkerOptions}
        @param render_options: options for the renderer
        @type render_options: L{arcticzim.zimbuild.renderer.RenderOptions}
        """
        # prepare the process priority
        config_process(name="AZ worker {}".format(id), nice=10, ionice=5)
        # start the worker
        worker = Worker(
            id=id,
            inqueue=self.inqueue,
            outqueue=self.outqueue,
            engine=connection_config.connect(),
            options=worker_options,
            render_options=render_options,
        )
        worker.run()

    def _send_post_tasks(self, session):
        """
        Create and send the tasks for the posts to the worker inqueue.

        @param session: sqlalchemy session for data querying
        @type session: L{sqlalchemy.orm.Session}
        """
        post_bucket_maker = BucketMaker(maxsize=POSTS_PER_TASK)
        select_post_ids_stmt = select(Post.uid)
        result = session.execute(select_post_ids_stmt)
        # create buckets and turn them into tasks
        for post in result:
            bucket = post_bucket_maker.feed(post.uid)
            if bucket is not None:
                # send out a task
                task = PostRenderTask(post_uids=bucket)
                self.inqueue.put(task)
        # send out all remaining tasks
        bucket = post_bucket_maker.finish()
        if bucket is not None:
            task = PostRenderTask(bucket)
            self.inqueue.put(task)

    def _send_subreddit_tasks(self, session, with_stats=True):
        """
        Create and send the tasks for the subreddits to the worker inqueue.

        @param session: sqlalchemy session for data querying
        @type session: L{sqlalchemy.orm.Session}
        @param with_stats: if nonzero, include stats
        @type with_stats: L{bool}
        """
        select_subreddit_names_stmt = select(Subreddit.name)
        result = session.execute(select_subreddit_names_stmt)
        subtasks = ["top", "new", "wiki", "rules"]
        if with_stats:
            subtasks.append("stats")
        # send out tasks
        for subreddit in result:
            for subtask in subtasks:
                task = SubredditRenderTask(subreddit_name=subreddit.name, subtask=subtask)
                self.inqueue.put(task)

    def _send_user_tasks(self, session, with_stats=True):
        """
        Create and send the tasks for the users to the worker inqueue.

        @param session: sqlalchemy session for data querying
        @type session: L{sqlalchemy.orm.Session}
        @param with_stats: if nonzero, include stats
        @type with_stats: L{bool}
        """
        select_usernames_stmt = select(User.name).where(User.name != ARCTICZIM_USERNAME)
        result = session.execute(select_usernames_stmt)
        # send out tasks
        for user in result:
            for part in ("posts", "comments"):
                for sort in ("top", "new"):
                    task = UserRenderTask(username=user.name, part=part, sort=sort)
                    self.inqueue.put(task)
            if with_stats:
                statstask = UserRenderTask(username=user.name, part="stats", sort=None)
                self.inqueue.put(statstask)

    def _send_etc_tasks(self, session, with_stats=True):
        """
        Create and send the tasks for the miscelaneous pages to the worker inqueue.

        @param session: sqlalchemy session for data querying
        @type session: L{sqlalchemy.orm.Session}
        @param with_stats: if nonzero, include stats
        @type with_stats: L{bool}
        """
        tasknames = ["index", "scripts", "subreddits", "info"]
        if with_stats:
            tasknames.append("stats")
        for taskname in tasknames:
            task = EtcRenderTask(taskname)
            self.inqueue.put(task)



"""
This module handles the generation of statistics.
"""
from sqlalchemy import select, func, distinct

from ..db.models import Post, Comment


class PostListStatistics(object):
    """
    This class holds statistics about a list of posts.

    @ivar count: total amount of posts
    @type count: L{int}

    @ivar total_score: total combined score of all posts
    @type total_score: L{int}
    @ivar min_score: score of the lowest scored post
    @type min_score: L{int}
    @ivar max_score: score of the heighest scored post
    @type max_score: L{int}

    @ivar oldest_utc: creation time of the oldest post
    @type oldest_utc: L{int}
    @ivar newest_utc: creation time of the oldest post
    @type newest_utc: L{int}

    @ivar total_comments: total amount of comments across all posts
    @type total_comments: L{int}
    @ivar max_comments: lowest amount of comments in a post
    @type max_comments: L{int}
    @ivar min_comments: highest amount of comments in a post
    @type min_comments: L{int}

    @ivar num_posters: number of users who have submitted a post
    @type num_posters: L{int}
    @ivar num_commentors: number of users who have submitted a comment
    @type num_commentors: L{int}
    """
    def __init__(
        self,
        count,
        total_score,
        min_score,
        max_score,
        oldest_utc,
        newest_utc,
        total_comments,
        max_comments,
        min_comments,
        num_posters,
        num_commentors,
    ):
        """
        The default constructor.

        @param count: total amount of posts
        @type count: L{int}

        @param total_score: total combined score of all posts
        @type total_score: L{int}
        @param min_score: score of the lowest scored post
        @type min_score: L{int}
        @param max_score: score of the heighest scored post
        @type max_score: L{int}

        @param oldest_utc: creation time of the oldest post
        @type oldest_utc: L{int}
        @param newest_utc: creation time of the oldest post
        @type newest_utc: L{int}

        @param total_comments: total amount of comments across all posts
        @type total_comments: L{int}
        @param max_comments: lowest amount of comments in a post
        @type max_comments: L{int}
        @param min_comments: highest amount of comments in a post
        @type min_comments: L{int}

        @param num_posters: number of users who have submitted a post
        @type num_posters: L{int}
        @param num_commentors: number of users who have submitted a comment
        @type num_commentors: L{int}
        """
        self.count = count
        self.total_score = total_score
        self.min_score = min_score
        self.max_score = max_score
        self.oldest_utc = oldest_utc
        self.newest_utc = newest_utc
        self.total_comments = total_comments
        self.max_comments = max_comments
        self.min_comments = min_comments
        self.num_posters = num_posters
        self.num_commentors = num_commentors

    @property
    def average_score(self):
        """
        Calculate the average score of the posts.

        @return: the average score of the posts
        @rtype: L{float} or L{None}
        """
        if self.count == 0:
            return None
        return self.total_score / self.count

    @property
    def average_comments(self):
        """
        Calculate the average amount of comments of the posts.

        @return: the average amount of comments of the posts
        @rtype: L{float} or L{None}
        """
        if self.count == 0:
            return None
        return self.total_comments / self.count

    @property
    def average_number_of_posts_per_poster(self):
        """
        Calculate the average amount of posts made by users who have submitted at least one post.

        @return: the average amount of posts made by users who have submitted at least one post.
        @rtype: L{float}
        """
        if self.num_posters == 0:
            return None
        return self.count / self.num_posters

    @property
    def average_number_of_comments_per_commentor(self):
        """
        Calculate the average amount of comments made by users who have submitted at least one comment.

        @return: the average amount of comment made by users who have submitted at least one comment.
        @rtype: L{float}
        """
        if self.num_commentors == 0:
            return None
        return self.total_comments / self.num_commentors


def query_post_stats(session, post_filter, comment_filter):
    """
    Query the statistics.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param post_filter: filter to apply on posts
    @type post_filter: sqlalchemy filter condition
    @param comment_filter: filter to apply on comments
    @type comment_filter: sqlalchemy filter condition
    @return: the collected statistics
    @rtype: L{PostListStatistics}
    """
    post_stat_stmt = select(
        func.count(Post.uid).label("count"),
        func.sum(Post.score).label("total_score"),
        func.min(Post.score).label("min_score"),
        func.max(Post.score).label("max_score"),
        func.min(Post.created_utc).label("oldest_utc"),
        func.max(Post.created_utc).label("newest_utc"),
        func.count(distinct(Post.author_name)).label("num_posters"),
    ).where(
        post_filter
    )
    post_result = session.execute(post_stat_stmt).one()
    comment_group_subq = select(
        Comment.link_id,
        func.count(Comment.uid).label("count"),
    ).where(
        comment_filter
    ).group_by(
        Comment.link_id,
    ).subquery()
    comment_stat_stmt = select(
        func.sum(comment_group_subq.c.count).label("total_comments"),
        func.min(comment_group_subq.c.count).label("min_comments"),
        func.max(comment_group_subq.c.count).label("max_comments"),
    )
    comment_result = session.execute(comment_stat_stmt).one()
    commentor_stmt = select(
        func.count(distinct(Comment.author_name)).label("num_commentors"),
    )
    commentor_result = session.execute(commentor_stmt).one()
    stats = PostListStatistics(
        count=post_result.count,
        total_score=post_result.total_score,
        min_score=post_result.min_score,
        max_score=post_result.max_score,
        oldest_utc=post_result.oldest_utc,
        newest_utc=post_result.newest_utc,
        total_comments=(comment_result.total_comments if comment_result.total_comments is not None else 0),
        min_comments=comment_result.min_comments,
        max_comments=comment_result.max_comments,
        num_posters=post_result.num_posters,
        num_commentors=commentor_result.num_commentors,
    )
    return stats
